from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass

import numpy as np
from pydantic import BaseModel, Field

from ragas.experimental.llms.prompt import PydanticPrompt
from ragas.metrics.base import EvaluationMode, MetricWithLLM, get_segmenter

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.metrics._faithfulness import HasSegmentMethod


logger = logging.getLogger(__name__)


class FaithfulnessStatements(BaseModel):
    question: str = Field(description="The question to answer")
    answer: str = Field(description="The answer to the question")
    sentences: t.Dict[int, str] = Field(
        description="A mapping of sentence index to the sentence"
    )


class SentenceComponents(BaseModel):
    sentence_index: int = Field(description="The index of the sentence")
    simpler_statements: t.List[str] = Field(
        description="A list of simpler statements that can be directly inferred from the context"
    )


class SentencesSimplified(BaseModel):
    sentences: t.List[SentenceComponents] = Field(
        description="A list of sentences and their simpler versions"
    )


# examples
example_input_1 = FaithfulnessStatements(
    question="Who was Albert Einstein and what is he best known for?",
    answer="He was a German-born theoretical physicist, widely acknowledged to be one of the greatest and most influential physicists of all time. He was best known for developing the theory of relativity, he also made important contributions to the development of the theory of quantum mechanics.",
    sentences={
        0: "He was a German-born theoretical physicist, widely acknowledged to be one of the greatest and most influential physicists of all time.",
        1: "He was best known for developing the theory of relativity, he also made important contributions to the development of the theory of quantum mechanics.",
    },
)

example_output_1 = SentencesSimplified(
    sentences=[
        SentenceComponents(
            sentence_index=0,
            simpler_statements=[
                "Albert Einstein was a German-born theoretical physicist.",
                "Albert Einstein is recognized as one of the greatest and most influential physicists of all time.",
            ],
        ),
        SentenceComponents(
            sentence_index=1,
            simpler_statements=[
                "Albert Einstein was best known for developing the theory of relativity.",
                "Albert Einstein also made important contributions to the development of the theory of quantum mechanics.",
            ],
        ),
    ]
)


class LongFormAnswerPrompt(PydanticPrompt[FaithfulnessStatements, SentencesSimplified]):
    instruction = "Given a question, an answer, and sentences from the answer analyze the complexity of each sentence given under 'sentences' and break down each sentence into one or more fully understandable statements while also ensuring no pronouns are used in each statement. Format the outputs in JSON."
    input_model = FaithfulnessStatements
    output_model = SentencesSimplified
    examples = [(example_input_1, example_output_1)]


class StatementFaithfulnessAnswer(BaseModel):
    statement: str = Field(..., description="the original statement, word-by-word")
    reason: str = Field(..., description="the reason of the verdict")
    verdict: int = Field(..., description="the verdict(0/1) of the faithfulness.")


class NLIStatementOutput(BaseModel):
    statements: t.List[StatementFaithfulnessAnswer]


class NLIStatementInput(BaseModel):
    context: str = Field(..., description="The context of the question")
    statements: t.List[str] = Field(..., description="The statements to judge")


class NLIStatementPrompt(PydanticPrompt[NLIStatementInput, NLIStatementOutput]):
    instruction = "Your task is to judge the faithfulness of a series of statements based on a given context. For each statement you must return verdict as 1 if the statement can be directly inferred based on the context or 0 if the statement can not be directly inferred based on the context."
    input_model = NLIStatementInput
    output_model = NLIStatementOutput
    examples = [
        (
            NLIStatementInput(
                context="""John is a student at XYZ University. He is pursuing a degree in Computer Science. He is enrolled in several courses this semester, including Data Structures, Algorithms, and Database Management. John is a diligent student and spends a significant amount of time studying and completing assignments. He often stays late in the library to work on his projects.""",
                statements=[
                    "John is majoring in Biology.",
                    "John is taking a course on Artificial Intelligence.",
                    "John is a dedicated student.",
                    "John has a part-time job.",
                ],
            ),
            NLIStatementOutput(
                statements=[
                    StatementFaithfulnessAnswer(
                        statement="John is majoring in Biology.",
                        reason="John's major is explicitly mentioned as Computer Science. There is no information suggesting he is majoring in Biology.",
                        verdict=0,
                    ),
                    StatementFaithfulnessAnswer(
                        statement="John is taking a course on Artificial Intelligence.",
                        reason="The context mentions the courses John is currently enrolled in, and Artificial Intelligence is not mentioned. Therefore, it cannot be deduced that John is taking a course on AI.",
                        verdict=0,
                    ),
                    StatementFaithfulnessAnswer(
                        statement="John is a dedicated student.",
                        reason="The context states that he spends a significant amount of time studying and completing assignments. Additionally, it mentions that he often stays late in the library to work on his projects, which implies dedication.",
                        verdict=1,
                    ),
                    StatementFaithfulnessAnswer(
                        statement="John has a part-time job.",
                        reason="There is no information given in the context about John having a part-time job.",
                        verdict=0,
                    ),
                ]
            ),
        ),
        (
            NLIStatementInput(
                context="Photosynthesis is a process used by plants, algae, and certain bacteria to convert light energy into chemical energy.",
                statements=[
                    "Albert Einstein was a genius.",
                ],
            ),
            NLIStatementOutput(
                statements=[
                    StatementFaithfulnessAnswer(
                        statement="Albert Einstein was a genius.",
                        reason="The context and statement are unrelated",
                        verdict=0,
                    )
                ]
            ),
        ),
    ]


@dataclass
class FaithfulnessExperimental(MetricWithLLM):
    name: str = "faithfulness_experimental"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    sentence_segmenter: t.Optional[HasSegmentMethod] = None
    max_retries: int = 1
    _reproducibility: int = 1

    @property
    def reproducibility(self):
        return self._reproducibility

    @reproducibility.setter
    def reproducibility(self, value):
        if value < 1:
            logger.warning("reproducibility cannot be less than 1, setting to 1")
            value = 1
        elif value % 2 == 0:
            logger.warning(
                "reproducibility level cannot be set to even number, setting to odd"
            )
            value += 1
        self._reproducibility = value

    def __post_init__(self):
        self.long_form_answer_prompt = LongFormAnswerPrompt(llm=self.llm)
        self.nli_statement_prompt = NLIStatementPrompt(llm=self.llm)
        if self.sentence_segmenter is None:
            # TODO: make this dynamic, taking language from prompt
            language = "english"
            self.sentence_segmenter = get_segmenter(language=language, clean=False)

    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        answer, question, contexts = row["answer"], row["question"], row["contexts"]

        # get the sentences from the answer
        if self.sentence_segmenter is None:
            raise ValueError("Sentence segmenter is not set")
        sentences = self.sentence_segmenter.segment(answer)
        # TODO: why do we do this?
        sentences = [
            sentence for sentence in sentences if sentence.strip().endswith(".")
        ]
        sentence_components = await self.long_form_answer_prompt.generate(
            FaithfulnessStatements(
                question=question,
                answer=answer,
                sentences={i: sentence for i, sentence in enumerate(sentences)},
            ),
            callbacks=callbacks,
        )

        statements = [
            statement
            for component in sentence_components.sentences
            for statement in component.simpler_statements
        ]
        verdicts = await self.nli_statement_prompt.generate(
            NLIStatementInput(
                context="\n".join(contexts),
                statements=statements,
            ),
            callbacks=callbacks,
        )

        # compute the score
        num_faithful_statements = sum(
            verdict.verdict for verdict in verdicts.statements
        )
        if len(statements):
            score = num_faithful_statements / len(statements)
        else:
            score = np.nan
        return score



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np
from langchain_core.pydantic_v1 import BaseModel

from ragas.llms.output_parser import RagasoutputParser, get_json_format_instructions
from ragas.llms.prompt import Prompt
from ragas.metrics.base import EvaluationMode, MetricWithLLM, ensembler

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.llms.prompt import PromptValue

logger = logging.getLogger(__name__)


class ContextRecallClassificationAnswer(BaseModel):
    statement: str
    attributed: int
    reason: str


class ContextRecallClassificationAnswers(BaseModel):
    __root__: t.List[ContextRecallClassificationAnswer]

    def dicts(self) -> t.List[t.Dict]:
        return self.dict()["__root__"]


_classification_output_instructions = get_json_format_instructions(
    ContextRecallClassificationAnswers
)
_output_parser = RagasoutputParser(pydantic_object=ContextRecallClassificationAnswers)


CONTEXT_RECALL_RA = Prompt(
    name="context_recall",
    instruction="""Given a context, and an answer, analyze each sentence in the answer and classify if the sentence can be attributed to the given context or not. Use only "Yes" (1) or "No" (0) as a binary classification. Output json with reason.""",
    output_format_instruction=_classification_output_instructions,
    examples=[
        {
            "question": """What can you tell me about albert Albert Einstein?""",
            "context": """Albert Einstein (14 March 1879 - 18 April 1955) was a German-born theoretical physicist, widely held to be one of the greatest and most influential scientists of all time. Best known for developing the theory of relativity, he also made important contributions to quantum mechanics, and was thus a central figure in the revolutionary reshaping of the scientific understanding of nature that modern physics accomplished in the first decades of the twentieth century. His mass-energy equivalence formula E = mc2, which arises from relativity theory, has been called 'the world's most famous equation'. He received the 1921 Nobel Prize in Physics 'for his services to theoretical physics, and especially for his discovery of the law of the photoelectric effect', a pivotal step in the development of quantum theory. His work is also known for its influence on the philosophy of science. In a 1999 poll of 130 leading physicists worldwide by the British journal Physics World, Einstein was ranked the greatest physicist of all time. His intellectual achievements and originality have made Einstein synonymous with genius.""",
            "answer": """Albert Einstein born in 14 March 1879 was  German-born theoretical physicist, widely held to be one of the greatest and most influential scientists of all time. He received the 1921 Nobel Prize in Physics for his services to theoretical physics. He published 4 papers in 1905.  Einstein moved to Switzerland in 1895""",
            "classification": ContextRecallClassificationAnswers.parse_obj(
                [
                    {
                        "statement": "Albert Einstein, born on 14 March 1879, was a German-born theoretical physicist, widely held to be one of the greatest and most influential scientists of all time.",
                        "reason": "The date of birth of Einstein is mentioned clearly in the context.",
                        "attributed": 1,
                    },
                    {
                        "statement": "He received the 1921 Nobel Prize in Physics for his services to theoretical physics.",
                        "reason": "The exact sentence is present in the given context.",
                        "attributed": 1,
                    },
                    {
                        "statement": "He published 4 papers in 1905.",
                        "reason": "There is no mention about papers he wrote in the given context.",
                        "attributed": 0,
                    },
                    {
                        "statement": "Einstein moved to Switzerland in 1895.",
                        "reason": "There is no supporting evidence for this in the given context.",
                        "attributed": 0,
                    },
                ]
            ).dicts(),
        },
        {
            "question": """who won 2020 icc world cup?""",
            "context": """The 2022 ICC Men's T20 World Cup, held from October 16 to November 13, 2022, in Australia, was the eighth edition of the tournament. Originally scheduled for 2020, it was postponed due to the COVID-19 pandemic. England emerged victorious, defeating Pakistan by five wickets in the final to clinch their second ICC Men's T20 World Cup title.""",
            "answer": """England""",
            "classification": ContextRecallClassificationAnswers.parse_obj(
                [
                    {
                        "statement": "England won the 2022 ICC Men's T20 World Cup.",
                        "reason": "From context it is clear that England defeated Pakistan to win the World Cup.",
                        "attributed": 1,
                    },
                ]
            ).dicts(),
        },
        {
            "question": """What is the primary fuel for the Sun?""",
            "context": """NULL""",
            "answer": """Hydrogen""",
            "classification": ContextRecallClassificationAnswers.parse_obj(
                [
                    {
                        "statement": "The Sun's primary fuel is hydrogen.",
                        "reason": "The context contains no information",
                        "attributed": 0,
                    },
                ]
            ).dicts(),
        },
    ],
    input_keys=["question", "context", "answer"],
    output_key="classification",
    output_type="json",
)


@dataclass
class ContextRecall(MetricWithLLM):
    """
    Estimates context recall by estimating TP and FN using annotated answer and
    retrieved context.

    Attributes
    ----------
    name : str
    """

    name: str = "context_recall"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qcg  # type: ignore
    context_recall_prompt: Prompt = field(default_factory=lambda: CONTEXT_RECALL_RA)
    max_retries: int = 1
    _reproducibility: int = 1

    @property
    def reproducibility(self):
        return self._reproducibility

    @reproducibility.setter
    def reproducibility(self, value):
        if value < 1:
            logger.warning("reproducibility cannot be less than 1, setting to 1")
            value = 1
        elif value % 2 == 0:
            logger.warning(
                "reproducibility level cannot be set to even number, setting to odd"
            )
            value += 1
        self._reproducibility = value

    def __post_init__(self) -> None:
        if self.reproducibility < 1:
            logger.warning("reproducibility cannot be less than 1, setting to 1")
            self.reproducibility = 1

    def _create_context_recall_prompt(self, row: t.Dict) -> PromptValue:
        qstn, ctx, gt = row["question"], row["contexts"], row["ground_truth"]
        ctx = "\n".join(ctx) if isinstance(ctx, list) else ctx

        return self.context_recall_prompt.format(question=qstn, context=ctx, answer=gt)

    def _compute_score(self, response: t.Any) -> float:
        response = [1 if item.attributed else 0 for item in response.__root__]
        denom = len(response)
        numerator = sum(response)
        score = numerator / denom if denom > 0 else np.nan

        if np.isnan(score):
            logger.warning("The LLM did not return a valid classification.")

        return score

    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "set LLM before use"
        p_value = self._create_context_recall_prompt(row)
        results = await self.llm.generate(
            p_value,
            callbacks=callbacks,
            n=self.reproducibility,
        )
        results = [results.generations[0][i].text for i in range(self.reproducibility)]

        answers = [
            await _output_parser.aparse(text, p_value, self.llm, self.max_retries)
            for text in results
        ]

        answers = [answer.dicts() for answer in answers if answer is not None]
        if all(answer is None for answer in answers):
            return np.nan

        answers = ensembler.from_discrete(answers, "attributed")
        answers = ContextRecallClassificationAnswers.parse_obj(answers)

        return self._compute_score(answers)

    def adapt(self, language: str, cache_dir: str | None = None) -> None:
        assert self.llm is not None, "set LLM before use"

        logger.info(f"Adapting Context Recall to {language}")
        self.context_recall_prompt = self.context_recall_prompt.adapt(
            language, self.llm, cache_dir
        )

    def save(self, cache_dir: str | None = None) -> None:
        self.context_recall_prompt.save(cache_dir)


context_recall = ContextRecall()



import json
import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np
import tiktoken
from langchain.utils.math import cosine_similarity
from langchain_core.documents import Document as LCDocument
from ragas_experimental.testset.graph import Node
from ragas_experimental.testset.questions.base import (
    QAC,
    QAGenerator,
    QuestionLength,
    QuestionStyle,
)
from ragas_experimental.testset.questions.prompts import (
    abstract_comparative_question,
    abstract_question_from_theme,
    common_theme_from_summaries,
    common_topic_from_keyphrases,
    critic_question,
    question_answering,
)
from ragas_experimental.testset.questions.queries import (
    CLUSTER_OF_RELATED_NODES_QUERY,
    LEAF_NODE_QUERY,
)
from ragas_experimental.testset.utils import rng

from ragas.executor import Executor
from ragas.llms.json_load import json_loader
from ragas.llms.prompt import Prompt

logger = logging.getLogger(__name__)


@dataclass
class AbstractQuestions(QAGenerator):
    def query_nodes(self, query: str, kwargs) -> t.Any:
        return super().query_nodes(query, kwargs)

    def filter_nodes(self, query: str, kwargs) -> t.Any:
        results = self.query_nodes(query, kwargs)
        sets = []
        for cluster in results:
            ids = [cluster.id]
            ids.extend([rel.target.id for rel in cluster.relationships])
            sets.append(set(ids))

        sets = np.unique(sets)
        # Convert tuples to sets for easy comparison
        sets_as_sets = [set(s) for s in sets]
        supersets = []
        for s in sets_as_sets:
            is_subset = False
            for other in sets_as_sets:
                if other != s and other.issuperset(s):
                    is_subset = True
                    break
            if not is_subset:
                supersets.append(s)

        nodes_to_return = []
        for subset in supersets:
            nodes_to_return.append([node for node in self.nodes if node.id in subset])

        return nodes_to_return or None

    async def critic_question(self, question: str):
        raise NotImplementedError("critic_question is not implemented")

    async def generate_answer(self, question: str, chunks: t.List[LCDocument]):
        raise NotImplementedError("generate_answer is not implemented")

    async def generate_question(
        self,
        nodes: t.List[Node],
        style: QuestionStyle,
        length: QuestionLength,
        kwargs: t.Optional[dict] = None,
    ) -> t.Any:
        pass
        raise NotImplementedError("generate_question is not implemented")


@dataclass
class AbstractQA(AbstractQuestions):
    name: str = "AbstractQA"
    generate_question_prompt: Prompt = field(
        default_factory=lambda: abstract_question_from_theme
    )
    critic_question_prompt: Prompt = field(default_factory=lambda: critic_question)
    generate_answer_prompt: Prompt = field(default_factory=lambda: question_answering)
    generate_common_theme_prompt: Prompt = field(
        default_factory=lambda: common_theme_from_summaries
    )

    async def generate_questions(self, query, kwargs, num_samples=5):
        assert self.llm is not None, "LLM is not initialized"
        query = query or CLUSTER_OF_RELATED_NODES_QUERY
        if kwargs is None:
            kwargs = {
                "label": "summary_similarity",
                "property": "score",
                "value": 0.5,
                "comparison": "gt",
            }

        node_clusters = self.filter_nodes(query, kwargs)
        if node_clusters is None:
            logging.warning("No nodes satisfy the query")
            return QAC()

        num_clusters = min(num_samples, len(node_clusters))
        seed_per_results = num_samples // len(node_clusters)
        reminder = num_samples - seed_per_results * num_clusters
        seeds = [seed_per_results] * num_clusters
        seeds[-1] += reminder

        nodes_themes = []
        for cluster, num_seeds in zip(node_clusters, seeds):
            summaries = [item.properties["metadata"]["summary"] for item in cluster]
            summaries = "\n".join(
                [f"summary {i+1}: {summary}" for i, summary in enumerate(summaries)]
            )
            common_themes = await self.llm.generate(
                self.generate_common_theme_prompt.format(
                    summaries=summaries, num_themes=num_seeds
                )
            )
            common_themes = await json_loader.safe_load(
                common_themes.generations[0][0].text, llm=self.llm
            )
            nodes_themes.extend([(cluster, theme) for theme in common_themes])

        exec = Executor(
            desc="Generating",
            keep_progress_bar=True,
            raise_exceptions=True,
            run_config=None,
        )

        index = 0
        for dist, prob in self.distribution.items():
            style, length = dist
            for i in range(int(prob * num_samples)):
                exec.submit(
                    self.generate_question,
                    nodes_themes[index][0],
                    style,
                    length,
                    {"common_theme": nodes_themes[index][1]},
                )
                index += 1

        remaining_size = num_samples - index
        if remaining_size != 0:
            choices = np.array(self.distribution.keys())
            prob = np.array(self.distribution.values())
            random_distribution = rng.choice(choices, p=prob, size=remaining_size)
            for dist in random_distribution:
                style, length = dist
                exec.submit(
                    self.generate_question,
                    nodes_themes[index][0],
                    style,
                    length,
                    {"common_theme": nodes_themes[index][1]},
                )
                index += 1

        return exec.results()

    async def generate_question(
        self, nodes, style, length, kwargs: t.Optional[dict] = None
    ) -> QAC:
        assert self.llm is not None, "LLM is not initialized"
        kwargs = kwargs or {}
        common_theme = kwargs.get("common_theme", "")
        try:
            source = await self.retrieve_chunks(
                nodes, {"max_tokens": 4024, "topic": common_theme.get("description")}
            )
            if source:
                source_text = "\n\n".join([chunk.page_content for chunk in source])
                abstract_question = await self.llm.generate(
                    self.generate_question_prompt.format(
                        theme=common_theme["theme"], context=source_text
                    )
                )
                abstract_question = abstract_question.generations[0][0].text
                critic_verdict = await self.critic_question(abstract_question)
                if critic_verdict:
                    abstract_question = await self.modify_question(
                        abstract_question, style, length
                    )
                    answer = await self.generate_answer(abstract_question, source)
                    return QAC(
                        question=abstract_question,
                        answer=answer,
                        source=source,
                        name=self.name,
                        style=style,
                        length=length,
                    )
                else:
                    logger.warning(
                        "Critic rejected the question: %s", abstract_question
                    )
                    return QAC()
            else:
                logger.warning("source was not detected %s", common_theme)
                return QAC()

        except Exception as e:
            logger.error("Error while generating question: %s", e)
            raise e

    async def critic_question(self, question: str) -> bool:
        assert self.llm is not None, "LLM is not initialized"

        output = await self.llm.generate(critic_question.format(question=question))
        output = json.loads(output.generations[0][0].text)
        return all(score >= 2 for score in output.values())

    async def retrieve_chunks(
        self, nodes: t.List[Node], kwargs: t.Optional[dict] = None
    ) -> t.List[LCDocument] | None:
        assert self.embedding is not None, "Embedding is not initialized"
        assert self.llm is not None, "LLLM is not initialized"

        kwargs = kwargs or {}
        max_tokens = kwargs.get("max_tokens", 3000)
        topic = kwargs.get("topic", "")
        node_ids = [node.id for node in nodes]

        query = LEAF_NODE_QUERY
        leaf_nodes = [
            self.query_nodes(query, {"id": json.dumps(id)}) for id in node_ids
        ]
        leaf_nodes = [node for nodes in leaf_nodes for node in nodes]
        if leaf_nodes is None:
            return None

        output_documents = []
        # TODO: also filter summaries using cosine
        summaries = [
            f"{node.properties['metadata']['title']}: {node.properties['metadata']['summary']}"
            for node in nodes
        ]
        output_documents.append(
            LCDocument(
                page_content="\n".join(summaries), metadata={"source": "summary"}
            )
        )

        page_embeddings = [
            node.properties["metadata"]["page_content_embedding"] for node in leaf_nodes
        ]

        topic_embedding = await self.embedding.embed_text(topic)

        # TODO: replace with similarity top-k with threshold cutoff
        # print(cosine_similarity_top_k([topic_embedding], page_embeddings, score_threshold=0.7))
        similarity_matrix = cosine_similarity([topic_embedding], page_embeddings)
        most_similar = np.flip(np.argsort(similarity_matrix[0]))
        ranked_lead_nodes = [leaf_nodes[i] for i in most_similar]
        # TODO: allow for different models
        model_name = "gpt-3.5-turbo-"
        enc = tiktoken.encoding_for_model(model_name)
        ranked_chunks_length = [
            len(enc.encode(node.properties["page_content"]))
            for node in ranked_lead_nodes
        ]
        ranked_chunks_length = np.cumsum(ranked_chunks_length)
        index = np.argmax(np.argwhere(np.cumsum(ranked_chunks_length) < max_tokens)) + 1
        top_leaf_nodes = ranked_lead_nodes[:index]
        dict = {}

        for node in top_leaf_nodes:
            if node.properties["metadata"]["source"] in dict:
                dict[node.properties["metadata"]["source"]].append(
                    {node.level.value: node}
                )
            else:
                dict[node.properties["metadata"]["source"]] = [{node.level.value: node}]

        for source, level_to_node in dict.items():
            sorted_nodes = sorted(level_to_node, key=lambda x: list(x.keys())[0])
            dict[source] = sorted_nodes
            current_nodes = [elem for item in sorted_nodes for elem in item.values()]
            title, source_file = (
                current_nodes[0].properties["metadata"].get("title"),
                current_nodes[0].properties["metadata"].get("source"),
            )
            text = "\n\n".join(
                [node.properties["page_content"] for node in current_nodes]
            )
            text = f"Document title: {title}\n\n{text}"
            output_documents.append(
                LCDocument(page_content=text, metadata={"source": source_file})
            )

        return output_documents

    async def generate_answer(
        self,
        question: str,
        chunks: t.List[LCDocument],
    ) -> str:
        assert self.llm is not None, "LLM is not initialized"
        # TODO : add title+summary of each node + title + content from most relevant chunk
        text = "\n\n".join([chunk.page_content for chunk in chunks])
        output = await self.llm.generate(
            self.generate_answer_prompt.format(question=question, text=text)
        )
        return output.generations[0][0].text


@dataclass
class ComparativeAbstractQA(AbstractQuestions):
    name: str = "ComparitiveAbtractQA"
    common_topic_prompt: Prompt = field(
        default_factory=lambda: common_topic_from_keyphrases
    )
    generate_question_prompt: Prompt = field(
        default_factory=lambda: abstract_comparative_question
    )
    generate_answer_prompt: Prompt = field(default_factory=lambda: question_answering)

    async def generate_questions(self, query, kwargs, num_samples=5):
        assert self.llm is not None, "LLM is not initialized"
        query = query or CLUSTER_OF_RELATED_NODES_QUERY
        if kwargs is None:
            kwargs = {
                "label": "summary_similarity",
                "property": "score",
                "value": 0.5,
                "comparison": "gt",
            }

        node_clusters = self.filter_nodes(query, kwargs)
        if node_clusters is None:
            logging.warning("No nodes satisfy the query")
            return QAC()

        num_clusters = min(num_samples, len(node_clusters))
        seed_per_results = num_samples // len(node_clusters)
        reminder = num_samples - seed_per_results * num_clusters
        seeds = [seed_per_results] * num_clusters
        seeds[-1] += reminder

        nodes_themes = []
        for cluster, num_seeds in zip(node_clusters, seeds):
            keyphrases = [node.properties["metadata"]["keyphrases"] for node in cluster]

            common_themes = await self.llm.generate(
                self.common_topic_prompt.format(
                    keyphrases=keyphrases, num_concepts=num_seeds
                )
            )
            common_themes = await json_loader.safe_load(
                common_themes.generations[0][0].text, llm=self.llm
            )
            nodes_themes.extend([(cluster, theme) for theme in common_themes])

        exec = Executor(
            desc="Generating",
            keep_progress_bar=True,
            raise_exceptions=True,
            run_config=None,
        )

        index = 0
        for dist, prob in self.distribution.items():
            style, length = dist
            for i in range(int(prob * num_samples)):
                exec.submit(
                    self.generate_question,
                    nodes_themes[index][0],
                    style,
                    length,
                    {"common_theme": nodes_themes[index][1]},
                )
                index += 1

        remaining_size = num_samples - index
        if remaining_size != 0:
            choices = np.array(self.distribution.keys())
            prob = np.array(self.distribution.values())
            random_distribution = rng.choice(choices, p=prob, size=remaining_size)
            for dist in random_distribution:
                style, length = dist
                exec.submit(
                    self.generate_question,
                    nodes_themes[index][0],
                    style,
                    length,
                    {"common_theme": nodes_themes[index][1]},
                )
                index += 1

        return exec.results()

    async def generate_question(
        self, nodes, style, length, kwargs: t.Optional[dict] = None
    ) -> QAC:
        assert self.llm is not None, "LLM is not initialized"

        kwargs = kwargs or {}
        themes_and_keyphrases = kwargs.get("common_theme", {})
        common_themes = np.array(list(themes_and_keyphrases.keys()))
        # TODO: implement a better way to select the comparison topic
        selected_theme = rng.choice(common_themes, size=1)[0]
        keyphrases = themes_and_keyphrases[selected_theme]

        try:
            kwargs = {
                "max_tokens": 4024,
                "keyphrases": keyphrases,
                "common_theme": selected_theme,
            }
            source = await self.retrieve_chunks(nodes, kwargs)
            if source:
                question = await self.llm.generate(
                    self.generate_question_prompt.format(
                        concept=selected_theme,
                        keyphrases=keyphrases,
                        summaries=source[0].page_content,
                    )
                )
                question = question.generations[0][0].text
                critic = await self.critic_question(question)
                if critic:
                    question = await self.modify_question(question, style, length)
                    answer = await self.generate_answer(question, source)
                    return QAC(
                        question=question,
                        answer=answer,
                        source=source,
                        name=self.name,
                        style=style,
                        length=length,
                    )
                else:
                    logger.warning("question failed critic %s", question)
                    return QAC()
            else:
                logger.warning("source was not detected %s", selected_theme)
                return QAC()
        except Exception as e:
            logger.error("Error while generating question: %s", e)
            raise e

    async def critic_question(self, question: str) -> bool:
        assert self.llm is not None, "LLM is not initialized"

        output = await self.llm.generate(critic_question.format(question=question))
        output = json.loads(output.generations[0][0].text)
        return all(score >= 2 for score in output.values())

    async def generate_answer(self, question: str, chunks: t.List[LCDocument]) -> str:
        assert self.llm is not None, "LLM is not initialized"

        text = "\n\n".join([chunk.page_content for chunk in chunks])
        output = await self.llm.generate(
            self.generate_answer_prompt.format(question=question, text=text)
        )
        return output.generations[0][0].text

    async def retrieve_chunks(
        self, nodes: t.List[Node], kwargs: t.Optional[dict] = None
    ) -> t.List[LCDocument] | None:
        kwargs = kwargs or {}
        assert self.embedding is not None, "Embedding is not initialized"

        common_keyphrases = kwargs.get("keyphrases", [])
        common_theme = kwargs.get("common_theme", None)
        if common_theme is None:
            logging.warning("No common theme detected")
            return None
        max_tokens = kwargs.get("max_tokens", 4000)
        node_ids = [node.id for node in nodes]

        query = LEAF_NODE_QUERY
        leaf_nodes = [
            self.query_nodes(query, {"id": json.dumps(id)}) for id in node_ids
        ]
        leaf_nodes = [node for nodes in leaf_nodes for node in nodes]
        leaf_nodes = [
            node
            for node in leaf_nodes
            if any(
                phrase in node.properties["metadata"]["keyphrases"]
                for phrase in common_keyphrases
            )
        ]

        if leaf_nodes is None:
            logging.warning("No leaf nodes with given keyphrases")
            return None

        output_documents = []
        summaries = [
            f"{node.properties['metadata']['title']}: {node.properties['metadata']['summary']}"
            for node in nodes
        ]
        output_documents.append(
            LCDocument(
                page_content="\n".join(summaries), metadata={"source": "summary"}
            )
        )

        page_embeddings = [
            node.properties["metadata"]["page_content_embedding"] for node in leaf_nodes
        ]
        common_theme_embedding = await self.embedding.embed_text(common_theme)
        similarity_matrix = cosine_similarity([common_theme_embedding], page_embeddings)
        most_similar = np.flip(np.argsort(similarity_matrix[0]))
        ranked_lead_nodes = [leaf_nodes[i] for i in most_similar]
        # TODO: allow for different models
        model_name = "gpt-3.5-turbo-"
        enc = tiktoken.encoding_for_model(model_name)
        ranked_chunks_length = [
            len(enc.encode(node.properties["page_content"]))
            for node in ranked_lead_nodes
        ]
        ranked_chunks_length = np.cumsum(ranked_chunks_length)
        index = np.argmax(np.argwhere(np.cumsum(ranked_chunks_length) < max_tokens)) + 1
        top_leaf_nodes = ranked_lead_nodes[:index]
        dict = {}

        for node in top_leaf_nodes:
            if node.properties["metadata"]["source"] in dict:
                dict[node.properties["metadata"]["source"]].append(
                    {node.level.value: node}
                )
            else:
                dict[node.properties["metadata"]["source"]] = [{node.level.value: node}]

        for source, level_to_node in dict.items():
            sorted_nodes = sorted(level_to_node, key=lambda x: list(x.keys())[0])
            dict[source] = sorted_nodes
            current_nodes = [elem for item in sorted_nodes for elem in item.values()]
            title, source_file = (
                current_nodes[0].properties["metadata"].get("title"),
                current_nodes[0].properties["metadata"].get("source"),
            )
            text = "\n\n".join(
                [node.properties["page_content"] for node in current_nodes]
            )
            text = f"Document title: {title}\n\n{text}"
            output_documents.append(
                LCDocument(page_content=text, metadata={"source": source_file})
            )

        return output_documents



from ragas_experimental.testset.generators.base import QADistribution, TestGenerator
from ragas_experimental.testset.generators.simple import SimpleTestGenerator

__all__ = [
    "TestGenerator",
    "SimpleTestGenerator",
    "QADistribution",
]



from __future__ import annotations

import typing as t

from langchain.chains.base import Chain
from langchain.schema import RUN_KEY
from langchain_openai.chat_models import ChatOpenAI
from langchain_openai.embeddings import OpenAIEmbeddings
from langsmith.evaluation import EvaluationResult, RunEvaluator
from langsmith.schemas import Example, Run

from ragas.embeddings import LangchainEmbeddingsWrapper
from ragas.llms import LangchainLLMWrapper
from ragas.metrics.base import (
    EvaluationMode,
    Metric,
    MetricWithEmbeddings,
    MetricWithLLM,
    get_required_columns,
)
from ragas.run_config import RunConfig
from ragas.utils import get_or_init
from ragas.validation import EVALMODE_TO_COLUMNS

if t.TYPE_CHECKING:
    from langchain.callbacks.manager import (
        AsyncCallbackManagerForChainRun,
        CallbackManagerForChainRun,
    )


class EvaluatorChain(Chain, RunEvaluator):
    """
    Wrapper around ragas Metrics to use them with langsmith.
    """

    metric: Metric

    def __init__(self, metric: Metric, **kwargs: t.Any):
        kwargs["metric"] = metric
        super().__init__(**kwargs)
        if "run_config" in kwargs:
            run_config = kwargs["run_config"]
        else:
            run_config = RunConfig()
        if isinstance(self.metric, MetricWithLLM):
            llm = get_or_init(kwargs, "llm", ChatOpenAI)
            t.cast(MetricWithLLM, self.metric).llm = LangchainLLMWrapper(llm)
        if isinstance(self.metric, MetricWithEmbeddings):
            embeddings = get_or_init(kwargs, "embeddings", OpenAIEmbeddings)
            t.cast(
                MetricWithEmbeddings, self.metric
            ).embeddings = LangchainEmbeddingsWrapper(embeddings)
        self.metric.init(run_config)

    @property
    def input_keys(self) -> list[str]:
        return get_required_columns(self.metric.evaluation_mode)

    @property
    def output_keys(self) -> list[str]:
        return [self.metric.name]

    def _call(
        self,
        inputs: dict[str, t.Any],
        run_manager: t.Optional[CallbackManagerForChainRun] = None,
    ) -> dict[str, t.Any]:
        """
        Call the evaluation chain.
        """
        self._validate(inputs)
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()

        c = inputs.get("contexts", [""])
        g = inputs.get("ground_truth", "")
        q = inputs.get("question", "")
        a = inputs.get("answer", "")
        score = self.metric.score(
            {
                "question": q,
                "answer": a,
                "contexts": c,
                "ground_truth": g,
            },
            callbacks=callbacks,
        )
        return {self.metric.name: score}

    async def _acall(
        self,
        inputs: t.Dict[str, t.Any],
        run_manager: t.Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> t.Dict[str, t.Any]:
        """
        Call the evaluation chain.
        """
        self._validate(inputs)
        _run_manager = run_manager or AsyncCallbackManagerForChainRun.get_noop_manager()
        # TODO: currently AsyncCallbacks are not supported in ragas
        _run_manager.get_child()

        c = inputs.get("contexts", [""])
        g = inputs.get("ground_truth", "")
        q = inputs.get("question", "")
        a = inputs.get("answer", "")
        score = await self.metric.ascore(
            {
                "question": q,
                "answer": a,
                "contexts": c,
                "ground_truth": g,
            },
            callbacks=[],
        )
        return {self.metric.name: score}

    def _validate(
        self,
        input: dict[str, t.Any],
        question_key: str = "question",
        prediction_key: str = "answer",
        context_key: str = "contexts",
    ) -> None:
        # validate each example
        required_columns = EVALMODE_TO_COLUMNS[self.metric.evaluation_mode]
        if "question" in required_columns and question_key not in input:
            raise ValueError(
                f'"{question_key}" is required in each example'
                f"for the metric[{self.metric.name}] you have chosen."
            )
        if "answer" in required_columns and prediction_key not in input:
            raise ValueError(
                f'"{prediction_key}" is required in each prediction'
                f"for the metric[{self.metric.name}] you have chosen."
            )
        if "contexts" in required_columns and context_key not in input:
            raise ValueError(
                f'"{context_key}" is required in each prediction for the '
                f"metric[{self.metric.name}] you have chosen."
            )
        if "ground_truth" in required_columns and "ground_truth" not in input:
            raise ValueError(
                f'"ground_truth" is required in each prediction for the '
                f"metric[{self.metric.name}] you have chosen."
            )

    @staticmethod
    def _keys_are_present(keys_to_check: list, dict_to_check: dict) -> list[str]:
        return [k for k in keys_to_check if k not in dict_to_check]

    def _validate_langsmith_eval(self, run: Run, example: t.Optional[Example]) -> None:
        if example is None:
            raise ValueError(
                "expected example to be provided. Please check langsmith dataset and ensure valid dataset is uploaded."
            )
        if example.inputs is None:
            raise ValueError(
                "expected example.inputs to be provided. Please check langsmith dataset and ensure valid dataset is uploaded."
            )
        if example.outputs is None:
            raise ValueError(
                "expected example.inputs to be provided. Please check langsmith dataset and ensure valid dataset is uploaded."
            )
        if "question" not in example.inputs or "ground_truth" not in example.outputs:
            raise ValueError(
                "Expected 'question' and 'ground_truth' in example."
                f"Got: {[k for k in example.inputs.keys()]}"
            )
        assert (
            run.outputs is not None
        ), "the current run has no outputs. The chain should output 'answer' and 'contexts' keys."
        output_keys = get_required_columns(
            self.metric.evaluation_mode, ["question", "ground_truth"]
        )
        missing_keys = self._keys_are_present(output_keys, run.outputs)
        if missing_keys:
            raise ValueError(
                "Expected 'answer' and 'contexts' in run.outputs."
                f"Got: {[k for k in run.outputs.keys()]}"
            )

    def evaluate_run(
        self, run: Run, example: t.Optional[Example] = None
    ) -> EvaluationResult:
        """
        Evaluate a langsmith run
        """
        self._validate_langsmith_eval(run, example)

        # this is just to suppress the type checker error
        # actual check and error message is in the _validate_langsmith_eval
        assert run.outputs is not None
        assert example is not None
        assert example.inputs is not None
        assert example.outputs is not None

        chain_eval = run.outputs
        chain_eval["question"] = example.inputs["question"]
        if self.metric.evaluation_mode in [
            EvaluationMode.gc,
            EvaluationMode.ga,
            EvaluationMode.qcg,
            EvaluationMode.qga,
        ]:
            if example.outputs is None or "ground_truth" not in example.outputs:
                raise ValueError("expected `ground_truth` in example outputs.")
            chain_eval["ground_truth"] = example.outputs["ground_truth"]
        eval_output = self.invoke(chain_eval, include_run_info=True)

        evaluation_result = EvaluationResult(
            key=self.metric.name, score=eval_output[self.metric.name]
        )
        if RUN_KEY in eval_output:
            evaluation_result.evaluator_info[RUN_KEY] = eval_output[RUN_KEY]
        return evaluation_result



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod

import pydantic

# Check Pydantic version
from pydantic import BaseModel

from ragas.llms.output_parser import RagasoutputParser
from ragas.llms.prompt import PromptValue

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.llms.base import BaseRagasLLM

PYDANTIC_V2 = pydantic.VERSION.startswith("2.")


class BasePrompt(ABC):
    def __init__(self, llm):
        self.llm: BaseRagasLLM = llm

    @abstractmethod
    async def generate(self, data: t.Any, callbacks: Callbacks = None) -> t.Any:
        pass


def model_to_dict(
    model: BaseModel,
    by_alias: bool = False,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
) -> t.Dict[str, t.Any]:
    if PYDANTIC_V2:
        return model.model_dump(  # type: ignore
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )
    else:
        return model.dict(
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
        )


def to_json(model: t.Any, indent: int = 4) -> str:
    if PYDANTIC_V2:
        # Pydantic 2.x
        return model.model_dump_json(indent=indent)
    else:
        # Pydantic 1.x
        return model.json(indent=indent)


def model_to_json_schema(model: t.Type[BaseModel]) -> str:
    if PYDANTIC_V2:
        # NOTE: this is not the same as model.schema_json()
        return model.model_json_schema()  # type: ignore
    else:
        return model.schema_json()


InputModel = t.TypeVar("InputModel", bound=BaseModel)
OutputModel = t.TypeVar("OutputModel", bound=BaseModel)


class StringIO(BaseModel):
    text: str


class PydanticPrompt(BasePrompt, t.Generic[InputModel, OutputModel]):
    input_model: t.Type[InputModel]
    output_model: t.Type[OutputModel]
    instruction: str
    examples: t.List[t.Tuple[InputModel, OutputModel]] = []

    def generate_instruction(self) -> str:
        return self.instruction

    def generate_output_signature(self, indent: int = 4) -> str:
        schema = model_to_json_schema(self.output_model)
        return (
            f"Please return the output in a JSON format that complies with the "
            f"following schema as specified in JSON Schema and OpenAPI specification:\n"
            f"{schema}"
        )

    def generate_examples(self):
        if self.examples:
            example_strings = []
            for e in self.examples:
                input_data, output_data = e
                example_strings.append(
                    self.instruction
                    + "\n"
                    + "input: "
                    + to_json(input_data, indent=4)
                    + "\n"
                    + "output: "
                    + to_json(output_data, indent=4)
                )

            return (
                "These are some examples to show how to perform the above instruction\n"
                + "\n\n".join(example_strings)
            )
        # if no examples are provided
        else:
            return ""

    def to_string(self, data: InputModel) -> str:
        # this needs a check
        return (
            self.generate_instruction()
            + "\n"
            + self.generate_output_signature()
            + "\n"
            + self.generate_examples()
            + "\nNow perform the above instruction with the following input\n"
            + "input: "
            + to_json(data, indent=4)
            + "\n"
            + "output: "
        )

    async def generate(
        self, data: InputModel, callbacks: Callbacks = None
    ) -> OutputModel:
        prompt_value = PromptValue(prompt_str=self.to_string(data))
        resp = await self.llm.generate(prompt_value, callbacks=callbacks)
        resp_text = resp.generations[0][0].text
        parser = RagasoutputParser(pydantic_object=self.output_model)
        answer = await parser.aparse(resp_text, prompt_value, self.llm, max_retries=3)

        # TODO: make sure RagasOutputPraser returns the same type as OutputModel
        return answer  # type: ignore


class StringPrompt(BasePrompt):
    async def generate(self, data: str, callbacks: Callbacks = None) -> str:
        prompt_value = PromptValue(prompt_str=data)
        llm_result = await self.llm.agenerate_text(prompt_value, callbacks=callbacks)
        return llm_result.generations[0][0].text



from __future__ import annotations

import json
import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np
from langchain_core.pydantic_v1 import BaseModel, Field

from ragas.llms.output_parser import RagasoutputParser, get_json_format_instructions
from ragas.llms.prompt import Prompt
from ragas.metrics.base import EvaluationMode, MetricWithLLM, ensembler, get_segmenter

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.llms.prompt import PromptValue


class HasSegmentMethod(t.Protocol):
    def segment(self, text) -> t.Any:
        ...


logger = logging.getLogger(__name__)


class Statements(BaseModel):
    sentence_index: int = Field(
        ..., description="Index of the sentence from the statement list"
    )
    simpler_statements: t.List[str] = Field(..., description="the simpler statements")


class StatementsAnswers(BaseModel):
    __root__: t.List[Statements]

    def dicts(self) -> t.List[t.Dict]:
        return self.dict()["__root__"]


_statements_output_instructions = get_json_format_instructions(StatementsAnswers)
_statements_output_parser = RagasoutputParser(pydantic_object=StatementsAnswers)


LONG_FORM_ANSWER_PROMPT = Prompt(
    name="long_form_answer",
    output_format_instruction=_statements_output_instructions,
    instruction="Given a question, an answer, and sentences from the answer analyze the complexity of each sentence given under 'sentences' and break down each sentence into one or more fully understandable statements while also ensuring no pronouns are used in each statement. Format the outputs in JSON.",
    examples=[
        {
            "question": "Who was Albert Einstein and what is he best known for?",
            "answer": "He was a German-born theoretical physicist, widely acknowledged to be one of the greatest and most influential physicists of all time. He was best known for developing the theory of relativity, he also made important contributions to the development of the theory of quantum mechanics.",
            "sentences": """
        0:He was a German-born theoretical physicist, widely acknowledged to be one of the greatest and most influential physicists of all time. 
        1:He was best known for developing the theory of relativity, he also made important contributions to the development of the theory of quantum mechanics.
        """,
            "analysis": StatementsAnswers.parse_obj(
                [
                    {
                        "sentence_index": 0,
                        "simpler_statements": [
                            "Albert Einstein was a German-born theoretical physicist.",
                            "Albert Einstein is recognized as one of the greatest and most influential physicists of all time.",
                        ],
                    },
                    {
                        "sentence_index": 1,
                        "simpler_statements": [
                            "Albert Einstein was best known for developing the theory of relativity.",
                            "Albert Einstein also made important contributions to the development of the theory of quantum mechanics.",
                        ],
                    },
                ]
            ).dicts(),
        }
    ],
    input_keys=["question", "answer", "sentences"],
    output_key="analysis",
    language="english",
)


class StatementFaithfulnessAnswer(BaseModel):
    statement: str = Field(..., description="the original statement, word-by-word")
    reason: str = Field(..., description="the reason of the verdict")
    verdict: int = Field(..., description="the verdict(0/1) of the faithfulness.")


class StatementFaithfulnessAnswers(BaseModel):
    __root__: t.List[StatementFaithfulnessAnswer]

    def dicts(self) -> t.List[t.Dict]:
        return self.dict()["__root__"]


_faithfulness_output_instructions = get_json_format_instructions(
    StatementFaithfulnessAnswers
)
_faithfulness_output_parser = RagasoutputParser(
    pydantic_object=StatementFaithfulnessAnswers
)

NLI_STATEMENTS_MESSAGE = Prompt(
    name="nli_statements",
    instruction="Your task is to judge the faithfulness of a series of statements based on a given context. For each statement you must return verdict as 1 if the statement can be directly inferred based on the context or 0 if the statement can not be directly inferred based on the context.",
    output_format_instruction=_faithfulness_output_instructions,
    examples=[
        {
            "context": """John is a student at XYZ University. He is pursuing a degree in Computer Science. He is enrolled in several courses this semester, including Data Structures, Algorithms, and Database Management. John is a diligent student and spends a significant amount of time studying and completing assignments. He often stays late in the library to work on his projects.""",
            "statements": [
                "John is majoring in Biology.",
                "John is taking a course on Artificial Intelligence.",
                "John is a dedicated student.",
                "John has a part-time job.",
            ],
            "answer": StatementFaithfulnessAnswers.parse_obj(
                [
                    {
                        "statement": "John is majoring in Biology.",
                        "reason": "John's major is explicitly mentioned as Computer Science. There is no information suggesting he is majoring in Biology.",
                        "verdict": 0,
                    },
                    {
                        "statement": "John is taking a course on Artificial Intelligence.",
                        "reason": "The context mentions the courses John is currently enrolled in, and Artificial Intelligence is not mentioned. Therefore, it cannot be deduced that John is taking a course on AI.",
                        "verdict": 0,
                    },
                    {
                        "statement": "John is a dedicated student.",
                        "reason": "The context states that he spends a significant amount of time studying and completing assignments. Additionally, it mentions that he often stays late in the library to work on his projects, which implies dedication.",
                        "verdict": 1,
                    },
                    {
                        "statement": "John has a part-time job.",
                        "reason": "There is no information given in the context about John having a part-time job.",
                        "verdict": 0,
                    },
                ]
            ).dicts(),
        },
        {
            "context": """Photosynthesis is a process used by plants, algae, and certain bacteria to convert light energy into chemical energy.""",
            "statements": ["Albert Einstein was a genius."],
            "answer": StatementFaithfulnessAnswers.parse_obj(
                [
                    {
                        "statement": "Albert Einstein was a genius.",
                        "reason": "The context and statement are unrelated",
                        "verdict": 0,
                    }
                ]
            ).dicts(),
        },
    ],
    input_keys=["context", "statements"],
    output_key="answer",
    output_type="json",
    language="english",
)  # noqa: E501


@dataclass
class Faithfulness(MetricWithLLM):
    name: str = "faithfulness"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    nli_statements_message: Prompt = field(
        default_factory=lambda: NLI_STATEMENTS_MESSAGE
    )
    statement_prompt: Prompt = field(default_factory=lambda: LONG_FORM_ANSWER_PROMPT)
    sentence_segmenter: t.Optional[HasSegmentMethod] = None
    max_retries: int = 1
    _reproducibility: int = 1

    @property
    def reproducibility(self):
        return self._reproducibility

    @reproducibility.setter
    def reproducibility(self, value):
        if value < 1:
            logger.warning("reproducibility cannot be less than 1, setting to 1")
            value = 1
        elif value % 2 == 0:
            logger.warning(
                "reproducibility level cannot be set to even number, setting to odd"
            )
            value += 1
        self._reproducibility = value

    def __post_init__(self):
        if self.sentence_segmenter is None:
            language = self.nli_statements_message.language
            self.sentence_segmenter = get_segmenter(language=language, clean=False)

    def _create_nli_prompt(self, row: t.Dict, statements: t.List[str]) -> PromptValue:
        assert self.llm is not None, "llm must be set to compute score"

        contexts = row["contexts"]
        # check if the statements are support in the contexts
        contexts_str: str = "\n".join(contexts)
        statements_str: str = json.dumps(statements)
        prompt_value = self.nli_statements_message.format(
            context=contexts_str, statements=statements_str
        )
        return prompt_value

    def _create_statements_prompt(self, row: t.Dict) -> PromptValue:
        assert self.sentence_segmenter is not None, "sentence_segmenter is not set"

        answer, question = row["answer"], row["question"]
        sentences = self.sentence_segmenter.segment(answer)
        sentences = [
            sentence for sentence in sentences if sentence.strip().endswith(".")
        ]
        sentences = "\n".join([f"{i}:{x}" for i, x in enumerate(sentences)])
        prompt_value = self.statement_prompt.format(
            question=question, answer=answer, sentences=sentences
        )
        return prompt_value

    def _compute_score(self, answers: StatementFaithfulnessAnswers):
        # check the verdicts and compute the score
        faithful_statements = sum(
            1 if answer.verdict else 0 for answer in answers.__root__
        )
        num_statements = len(answers.__root__)
        if num_statements:
            score = faithful_statements / num_statements
        else:
            logger.warning("No statements were generated from the answer.")
            score = np.nan

        return score

    async def _ascore(self: t.Self, row: t.Dict, callbacks: Callbacks) -> float:
        """
        returns the NLI score for each (q, c, a) pair
        """
        assert self.llm is not None, "LLM is not set"

        p_value = self._create_statements_prompt(row)
        statements = await self.llm.generate(
            p_value,
            callbacks=callbacks,
        )
        statements = await _statements_output_parser.aparse(
            statements.generations[0][0].text, p_value, self.llm, self.max_retries
        )

        if statements is None:
            return np.nan

        statements = [item["simpler_statements"] for item in statements.dicts()]
        statements = [item for sublist in statements for item in sublist]

        assert isinstance(statements, t.List), "statements must be a list"

        p_value = self._create_nli_prompt(row, statements)
        nli_result = await self.llm.generate(
            p_value,
            callbacks=callbacks,
            n=self._reproducibility,
        )

        nli_result_text = [
            nli_result.generations[0][i].text for i in range(self._reproducibility)
        ]
        faithfulness_list = [
            await _faithfulness_output_parser.aparse(
                text, p_value, self.llm, self.max_retries
            )
            for text in nli_result_text
        ]

        faithfulness_list = [
            faith.dicts() for faith in faithfulness_list if faith is not None
        ]

        if faithfulness_list:
            faithfulness_list = ensembler.from_discrete(
                faithfulness_list,
                "verdict",
            )

            faithfulness_list = StatementFaithfulnessAnswers.parse_obj(
                faithfulness_list
            )
        else:
            return np.nan

        return self._compute_score(faithfulness_list)

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        assert self.llm is not None, "LLM is not set"

        logger.info(f"Adapting Faithfulness metric to {language}")

        self.nli_statements_message = self.nli_statements_message.adapt(
            language, self.llm, cache_dir
        )
        self.statement_prompt = self.statement_prompt.adapt(
            language, self.llm, cache_dir
        )

        self.sentence_segmenter = get_segmenter(language=language, clean=False)

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        self.nli_statements_message.save(cache_dir)
        self.statement_prompt.save(cache_dir)


@dataclass
class FaithulnesswithHHEM(Faithfulness):
    name: str = "faithfulness_with_hhem"  # type: ignore
    device: str = "cpu"
    batch_size: int = 10

    def __post_init__(self):
        try:
            from transformers import AutoModelForSequenceClassification
        except ImportError:
            raise ImportError(
                "Huggingface transformers must be installed to use this feature, try `pip install transformers`"
            )
        self.nli_classifier = AutoModelForSequenceClassification.from_pretrained(
            "vectara/hallucination_evaluation_model", trust_remote_code=True
        )
        self.nli_classifier.to(self.device)
        super().__post_init__()

    def _create_pairs(
        self, row: t.Dict, statements: t.List[str]
    ) -> t.List[t.Tuple[str, str]]:
        """
        create pairs of (question, answer) from the row
        """
        premise = "\n".join(row["contexts"])
        pairs = [(premise, statement) for statement in statements]
        return pairs

    def _create_batch(
        self, pairs: t.List[t.Tuple[str, str]]
    ) -> t.Generator[t.List[t.Tuple[str, str]], None, None]:
        length_of_pairs = len(pairs)
        for ndx in range(0, length_of_pairs, self.batch_size):
            yield pairs[ndx : min(ndx + self.batch_size, length_of_pairs)]

    async def _ascore(self: t.Self, row: t.Dict, callbacks: Callbacks) -> float:
        """
        returns the NLI score for each (q, c, a) pair
        """
        assert self.llm is not None, "LLM is not set"

        p_value = self._create_statements_prompt(row)
        statements = await self.llm.generate(
            p_value,
            callbacks=callbacks,
        )
        statements = await _statements_output_parser.aparse(
            statements.generations[0][0].text, p_value, self.llm, self.max_retries
        )

        if statements is None:
            return np.nan

        statements = [item["simpler_statements"] for item in statements.dicts()]
        statements = [item for sublist in statements for item in sublist]

        assert isinstance(statements, t.List), "statements must be a list"

        scores = []
        pairs = self._create_pairs(row, statements)
        for input_pairs in self._create_batch(pairs):  # to avoid OOM
            batch_scores = (
                self.nli_classifier.predict(input_pairs).cpu().detach().round()
            )
            scores += batch_scores
        return sum(scores) / len(scores)


faithfulness = Faithfulness()



from collections import namedtuple

import pytest
from datasets import Dataset

from ragas.metrics import answer_relevancy, context_precision, faithfulness
from ragas.validation import (
    remap_column_names,
    validate_column_dtypes,
    validate_evaluation_modes,
)

CaseToTest = namedtuple(
    "TestCase", ["q", "a", "c", "g", "is_valid_columns", "metrics", "is_valid_metrics"]
)

TEST_CASES = [
    CaseToTest("a", "b", ["c"], None, True, [faithfulness], True),
    CaseToTest("a", "b", ["c"], "g", True, [faithfulness], True),
    CaseToTest("a", None, ["c"], "g", True, [context_precision], True),
    CaseToTest("a", "b", "c", "g", False, [context_precision], True),
    CaseToTest(
        "a", None, [["c"]], None, False, [context_precision, answer_relevancy], False
    ),
    CaseToTest("a", None, ["c"], ["g"], False, [context_precision], True),
    CaseToTest("a", None, ["c"], [["g"]], False, [context_precision], True),
    CaseToTest(1, None, ["c"], "g", False, [context_precision], True),
    CaseToTest(1, None, None, None, False, [context_precision], False),
]


@pytest.mark.parametrize("testcase", TEST_CASES)
def test_validate_column_dtypes(testcase):
    dataset_dict = {}
    if testcase.q is not None:
        dataset_dict["question"] = [testcase.q]
    if testcase.a is not None:
        dataset_dict["answer"] = [testcase.a]
    if testcase.c is not None:
        dataset_dict["contexts"] = [testcase.c]
    if testcase.g is not None:
        dataset_dict["ground_truth"] = [testcase.g]

    test_dataset = Dataset.from_dict(dataset_dict)
    if testcase.is_valid_columns:
        validate_column_dtypes(test_dataset)
    else:
        with pytest.raises(ValueError):
            validate_column_dtypes(test_dataset)


@pytest.mark.parametrize("testcase", TEST_CASES)
def test_validate_columns_and_metrics(testcase):
    dataset_dict = {}
    if testcase.q is not None:
        dataset_dict["question"] = [testcase.q]
    if testcase.a is not None:
        dataset_dict["answer"] = [testcase.a]
    if testcase.c is not None:
        dataset_dict["contexts"] = [testcase.c]
    if testcase.g is not None:
        dataset_dict["ground_truth"] = [testcase.g]
    test_dataset = Dataset.from_dict(dataset_dict)

    if testcase.is_valid_metrics:
        validate_evaluation_modes(test_dataset, testcase.metrics)
    else:
        with pytest.raises(ValueError):
            validate_evaluation_modes(test_dataset, testcase.metrics)


column_maps = [
    {
        "question": "query",
        "answer": "rag_answer",
        "contexts": "rag_contexts",
        "ground_truth": "original_answer",
    },  # all columns present
    {
        "question": "query",
        "answer": "rag_answer",
    },  # subset of columns present
]


@pytest.mark.parametrize("column_map", column_maps)
def test_column_remap(column_map):
    """
    test cases:
    - extra columns present in the dataset
    - not all columsn selected
    - column names are different
    """
    TEST_DATASET = Dataset.from_dict(
        {
            "query": [""],
            "rag_answer": [""],
            "rag_contexts": [[""]],
            "original_answer": [""],
            "another_column": [""],
            "rag_answer_v2": [""],
            "rag_contexts_v2": [[""]],
        }
    )
    remapped_dataset = remap_column_names(TEST_DATASET, column_map)
    assert all(col in remapped_dataset.column_names for col in column_map.keys())


def test_column_remap_omit():
    TEST_DATASET = Dataset.from_dict(
        {
            "query": [""],
            "answer": [""],
            "contexts": [[""]],
        }
    )

    column_map = {
        "question": "query",
        "contexts": "contexts",
        "answer": "answer",
    }

    remapped_dataset = remap_column_names(TEST_DATASET, column_map)
    assert remapped_dataset.column_names == ["question", "answer", "contexts"]



from __future__ import annotations

import argparse
import asyncio
import os
import typing as t
from collections import namedtuple

from langchain.prompts import ChatPromptTemplate
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_openai.chat_models import ChatOpenAI
from tqdm.asyncio import tqdm

File = namedtuple("File", "name content")


def get_files(path: str, ext: str) -> list:
    return [os.path.join(path, f) for f in os.listdir(path) if f.endswith(ext)]


def load_docs(path: str) -> t.List[File]:
    files = [*get_files(path, ".md")]
    docs = []
    for file in files:
        with open(file, "r") as f:
            print("fixing: ", file)
            docs.append(File(file, f.read()))
    return docs


async def fix_doc_with_llm(doc: File, llm: BaseChatModel) -> File:
    prompt = """\
fix the following grammar and spelling mistakes in the following text. 
Please keep the markdown format intact when reformating it. 
Do not make any change to the parts of text that are for formating or additional metadata for the core text in markdown.
The target audience for this is developers so keep the tone serious and to the point without any marketing terms. 
The output text should me in .md format. 

text: {text}
"""
    fix_docs_prompt = ChatPromptTemplate.from_messages(
        [
            (prompt),
        ]
    )
    # get output
    fixed_doc = await llm.ainvoke(fix_docs_prompt.format_messages(text=doc.content))
    return File(doc.name, fixed_doc.content)


async def main(docs: t.List[File], llm: BaseChatModel):
    fix_doc_routines = [fix_doc_with_llm(doc, llm) for doc in docs]
    return await tqdm.gather(*fix_doc_routines)


if __name__ == "__main__":
    """
    Helpful assistant for documentation review and more (hopefully in the future).
    """
    # Create an argument parser
    parser = argparse.ArgumentParser(
        description="Helpful assistant for documentation review."
    )
    parser.add_argument("-d", "--directory", help="Directory to run the script against")
    args = parser.parse_args()
    directory = args.directory
    docs = load_docs(directory)
    gpt4 = ChatOpenAI(model="gpt-4")
    fix_docs = asyncio.run(main(docs, gpt4))
    for doc in fix_docs:
        with open(doc.name, "w") as f:
            f.write(doc.content)



import logging
import typing as t
from dataclasses import dataclass

from ragas_experimental.testset.graph import Node, NodeLevel, Relationship, schema
from ragas_experimental.testset.relationships.base import Similarity

logger = logging.getLogger(__name__)


@dataclass
class RelationshipBuilder:
    @staticmethod
    def form_relations(
        nodes: t.List[Node],
        relationships: t.List[Relationship],
        similarity_functions: t.List[Similarity],
        node_level: NodeLevel,
        kwargs: t.Optional[t.Dict[str, t.Any]] = None,
    ):
        kwargs = kwargs or {}
        score_threshold = kwargs.get("score_threshold", 0.0)

        query = """
            {{
            filterNodes(level: {node_level}) {{
                id
                label
                properties
                }}
            }}
                """.format(
            node_level=node_level.name
        )
        results = schema.execute(
            query, context={"nodes": nodes, "relationships": relationships}
        )

        # check if there is any data returned from the query
        if results.data is None:
            if results.errors:
                raise Exception(results.errors)
            logger.error("No data returned from query.")

        # since we already checked if there is any data returned from the query
        # to fix the type errors
        assert results.data, "No data returned from query."

        if not results.data.get("filterNodes"):
            logger.warning("No new relations created due to empty results.")
            return (nodes, relationships)

        node_ids = [item.get("id") for item in results.data["filterNodes"]]
        nodes_ = [node for node in nodes if node.id in node_ids]
        for extractor in similarity_functions:
            similarity_matrix = extractor.extract(nodes_, nodes_)
            for i, row in enumerate(similarity_matrix):
                new_relationships = []
                for j, score in enumerate(row):
                    if i != j and score >= score_threshold:
                        relationship = Relationship(
                            source=nodes_[i],
                            target=nodes_[j],
                            label=extractor.name,
                            properties={"score": score},
                        )
                        new_relationships.append(relationship)
                        relationship.source.relationships.append(relationship)
                        relationship.target.relationships.append(relationship)

                relationships.extend(new_relationships)

        return (nodes, relationships)



CLUSTER_OF_RELATED_NODES_QUERY = """
            {{
            filterNodes(label: DOC) {{
                id
                label
                properties
                relationships(label: "{label}", propertyKey: "{property}", propertyValue: "{value}", comparison: "{comparison}", targetFilter: {{label: DOC}} ) {{
                label
                properties
                target {{
                    id
                    label
                    properties
                }}
                }}
            }}
            }}
            """

LEAF_NODE_QUERY = """
{{
leafNodes(id: {id}){{
id
label
level
properties
}}
}}
"""

CHILD_NODES_QUERY = """
        {{
        filterNodes(label: DOC, level : LEVEL_0) {{
            id
            label
            properties
            relationships(label: "child") {{
            label
            properties
            target {{
                id
                label
                properties
            }}
            source {{
                id
            }}
            }}
        }}
        }}
        """



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field
from typing import Dict

from langchain.pydantic_v1 import BaseModel

from ragas.llms.output_parser import RagasoutputParser, get_json_format_instructions
from ragas.llms.prompt import Prompt
from ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks

logger = logging.getLogger(__name__)


class ExtractKeyphrasesResponse(BaseModel):
    keyphrases: t.List[str]


class GenerateQuestionsResponse(BaseModel):
    questions: t.List[str]


class GenerateAnswersResponse(BaseModel):
    answers: t.List[str]


_output_instructions_question_generation = get_json_format_instructions(
    pydantic_object=GenerateQuestionsResponse
)
_output_instructions_answer_generation = get_json_format_instructions(
    pydantic_object=GenerateAnswersResponse
)
_output_instructions_keyphrase_extraction = get_json_format_instructions(
    pydantic_object=ExtractKeyphrasesResponse
)
_output_parser_question_generation = RagasoutputParser(
    pydantic_object=GenerateQuestionsResponse
)
_output_parser_answer_generation = RagasoutputParser(
    pydantic_object=GenerateAnswersResponse
)
_output_parser_keyphrase_extraction = RagasoutputParser(
    pydantic_object=ExtractKeyphrasesResponse
)


TEXT_EXTRACT_KEYPHRASES = Prompt(
    name="text_extract_keyphrases",
    instruction="Extract the keyphrases essential for summarizing the text.",
    output_format_instruction=_output_instructions_keyphrase_extraction,
    input_keys=["text"],
    output_key="keyphrases",
    output_type="json",
    examples=[
        {
            "text": """JPMorgan Chase & Co. is an American multinational finance company headquartered in New York City. It is the largest bank in the United States and the world's largest by market capitalization as of 2023. Founded in 1799, it is a major provider of investment banking services, with US$3.9 trillion in total assets, and ranked #1 in the Forbes Global 2000 ranking in 2023.""",
            "keyphrases": [
                "JPMorgan Chase & Co.",
                "American multinational finance company",
                "headquartered in New York City",
                "largest bank in the United States",
                "world's largest bank by market capitalization",
                "founded in 1799",
                "major provider of investment banking services",
                "US$3.9 trillion in total assets",
                "ranked #1 in Forbes Global 2000 ranking",
            ],
        }
    ],
)


TEXT_GENERATE_QUESTIONS = Prompt(
    name="text_generate_questions",
    instruction="Based on the given text and keyphrases, generate closed-ended questions that can be answered with '1' if the question can be answered using the text, or '0' if it cannot. The questions should ALWAYS result in a '1' based on the given text.",
    output_format_instruction=_output_instructions_question_generation,
    input_keys=["text", "keyphrases"],
    output_key="questions",
    output_type="json",
    examples=[
        {
            "text": """JPMorgan Chase & Co. is an American multinational finance company headquartered in New York City. It is the largest bank in the United States and the world's largest by market capitalization as of 2023. Founded in 1799, it is a major provider of investment banking services, with US$3.9 trillion in total assets, and ranked #1 in the Forbes Global 2000 ranking in 2023.""",
            "keyphrases": [
                "JPMorgan Chase & Co.",
                "American multinational finance company",
                "headquartered in New York City",
                "largest bank in the United States",
                "world's largest bank by market capitalization",
                "founded in 1799",
                "major provider of investment banking services",
                "US$3.9 trillion in total assets",
                "ranked #1 in Forbes Global 2000 ranking",
            ],
            "questions": [
                "Is JPMorgan Chase & Co. an American multinational finance company?",
                "Is JPMorgan Chase & Co. headquartered in New York City?",
                "Is JPMorgan Chase & Co. the largest bank in the United States?",
                "Is JPMorgan Chase & Co. the world's largest bank by market capitalization as of 2023?",
                "Was JPMorgan Chase & Co. founded in 1799?",
                "Is JPMorgan Chase & Co. a major provider of investment banking services?",
                "Does JPMorgan Chase & Co. have US$3.9 trillion in total assets?",
                "Was JPMorgan Chase & Co. ranked #1 in the Forbes Global 2000 ranking in 2023?",
            ],
        }
    ],
)


TEXT_GENERATE_ANSWERS = Prompt(
    name="text_generate_answers",
    instruction="Based on the list of close-ended '1' or '0' questions, generate a JSON with key 'answers', which is a list of strings that determines whether the provided summary contains sufficient information to answer EACH question. Answers should STRICTLY be either '1' or '0'. Answer '0' if the provided summary does not contain enough information to answer the question and answer '1' if the provided summary can answer the question.",
    output_format_instruction=_output_instructions_answer_generation,
    input_keys=["summary", "questions"],
    output_key="answers",
    output_type="json",
    examples=[
        {
            "summary": """JPMorgan Chase & Co., headquartered in New York City, is the largest bank in the US and the world's largest by market capitalization as of 2023. Founded in 1799, it offers extensive investment, private, asset management, and retail banking services, and has $3.9 trillion in assets, making it the fifth-largest bank globally. It operates the world's largest investment bank by revenue and was ranked #1 in the 2023 Forbes Global 2000.""",
            "questions": [
                "Is JPMorgan Chase & Co. an American multinational finance company?",
                "Is JPMorgan Chase & Co. headquartered in New York City?",
                "Is JPMorgan Chase & Co. the largest bank in the United States?",
                "Is JPMorgan Chase & Co. the world's largest bank by market capitalization as of 2023?",
                "Is JPMorgan Chase & Co. considered systemically important by the Financial Stability Board?",
                "Was JPMorgan Chase & Co. founded in 1799 as the Chase Manhattan Company?",
                "Is JPMorgan Chase & Co. a major provider of investment banking services?",
                "Is JPMorgan Chase & Co. the fifth-largest bank in the world by assets?",
                "Does JPMorgan Chase & Co. operate the largest investment bank by revenue?",
                "Was JPMorgan Chase & Co. ranked #1 in the Forbes Global 2000 ranking?",
                "Does JPMorgan Chase & Co. provide investment banking services?",
            ],
            "answers": ["0", "1", "1", "1", "0", "0", "1", "1", "1", "1", "1"],
        }
    ],
)


@dataclass
class SummarizationScore(MetricWithLLM):
    name: str = "summary_score"  # type: ignore
    max_retries: int = 1
    length_penalty: bool = True
    coeff: float = 0.5
    evaluation_mode: EvaluationMode = EvaluationMode.ca  # type: ignore
    question_generation_prompt: Prompt = field(
        default_factory=lambda: TEXT_GENERATE_QUESTIONS
    )
    answer_generation_prompt: Prompt = field(
        default_factory=lambda: TEXT_GENERATE_ANSWERS
    )
    extract_keyphrases_prompt: Prompt = field(
        default_factory=lambda: TEXT_EXTRACT_KEYPHRASES
    )

    async def _ascore(self, row: Dict, callbacks: Callbacks) -> float:
        text: str = "\n".join(row["contexts"])
        summary: str = row["summary"]
        keyphrases = await self._extract_keyphrases(text, callbacks)
        questions = await self._get_questions(text, keyphrases, callbacks)
        answers = await self._get_answers(questions, summary, callbacks)

        scores = {}
        qa_score = self._compute_qa_score(answers)
        scores["qa_score"] = qa_score
        if self.length_penalty:
            conciseness_score = self._compute_conciseness_score(text, summary)
            scores["conciseness_score"] = conciseness_score
        return self._compute_score(scores)

    def _compute_score(self, scores) -> float:
        return (
            scores["qa_score"] * (1 - self.coeff)
            + scores.get("conciseness_score", 0) * self.coeff
        )

    def _compute_qa_score(self, answers: t.List[str]) -> float:
        correct = sum([1 for a in answers if a.lower() == "1"])
        return correct / len(answers)

    def _compute_conciseness_score(self, text, summary) -> float:
        return 1 - min(len(summary), len(text)) / (len(text) + 1e-10)

    async def _extract_keyphrases(self, text: str, callbacks: Callbacks) -> t.List[str]:
        assert self.llm is not None, "LLM is not initialized"
        p_value = self.extract_keyphrases_prompt.format(text=text)
        result = await self.llm.generate(
            prompt=p_value,
            callbacks=callbacks,
        )
        result_text = result.generations[0][0].text
        response = await _output_parser_keyphrase_extraction.aparse(
            result_text, p_value, self.llm, self.max_retries
        )

        if not response or not response.keyphrases:
            logging.error("No keyphrases generated, unable to calculate the score.")
            return []

        return response.keyphrases

    async def _get_questions(
        self, text: str, keyphrases: list[str], callbacks: Callbacks
    ) -> t.List[str]:
        assert self.llm is not None, "LLM is not initialized"
        p_value = self.question_generation_prompt.format(
            text=text, keyphrases=keyphrases
        )
        result = await self.llm.generate(
            prompt=p_value,
            callbacks=callbacks,
        )

        result_text = result.generations[0][0].text
        response = await _output_parser_question_generation.aparse(
            result_text, p_value, self.llm, self.max_retries
        )

        if not response or not response.questions:
            logging.error("No questions generated, unable to calculate the score.")
            return []

        return response.questions

    async def _get_answers(
        self, questions: t.List[str], summary: str, callbacks: Callbacks
    ) -> t.List[str]:
        assert self.llm is not None, "LLM is not initialized"
        p_value = self.answer_generation_prompt.format(
            questions=questions, summary=summary
        )
        result = await self.llm.generate(
            prompt=p_value,
            callbacks=callbacks,
        )

        result_text = result.generations[0][0].text
        response = await _output_parser_answer_generation.aparse(
            result_text, p_value, self.llm, self.max_retries
        )

        if not response or not response.answers:
            logger.error("No answers generated, unable to calculate the score.")
            return []

        return response.answers

    def adapt(self, language: str, cache_dir: str | None = None) -> None:
        assert self.llm is not None, "set LLM before use"

        logger.info(f"Adapting summarization to {language}")
        self.question_generation_prompt = self.question_generation_prompt.adapt(
            language, self.llm, cache_dir
        )
        self.answer_generation_prompt = self.answer_generation_prompt.adapt(
            language, self.llm, cache_dir
        )
        self.answer_generation_prompt = self.answer_generation_prompt.adapt(
            language, self.llm, cache_dir
        )


summarization_score = SummarizationScore()



import asyncio

import pytest


@pytest.mark.asyncio
async def test_order_of_execution():
    from ragas.executor import Executor

    async def echo_order(index: int):
        return index

    # Arrange
    executor = Executor()

    # Act
    # add 10 jobs to the executor
    for i in range(1, 11):
        executor.submit(echo_order, i, name=f"echo_order_{i}")
    results = executor.results()
    # Assert
    assert results == list(range(1, 11))


@pytest.mark.asyncio
async def test_executor_in_script():
    from ragas.executor import Executor

    async def echo_order(index: int):
        await asyncio.sleep(0.1)
        return index

    # Arrange
    executor = Executor()

    # Act
    # add 10 jobs to the executor
    for i in range(1, 4):
        executor.submit(echo_order, i, name=f"echo_order_{i}")
    results = executor.results()
    # Assert
    assert results == list(range(1, 4))


@pytest.mark.asyncio
async def test_executor_with_running_loop():
    import asyncio

    from ragas.executor import Executor

    loop = asyncio.new_event_loop()
    loop.run_until_complete(asyncio.sleep(0.1))

    async def echo_order(index: int):
        await asyncio.sleep(0.1)
        return index

    # Arrange
    executor = Executor()

    # Act
    # add 10 jobs to the executor
    for i in range(1, 4):
        executor.submit(echo_order, i, name=f"echo_order_{i}")
    results = executor.results()
    # Assert
    assert results == list(range(1, 4))


def test_is_event_loop_running_in_script():
    from ragas.executor import is_event_loop_running

    assert is_event_loop_running() is False


def test_as_completed_in_script():
    from ragas.executor import as_completed

    async def echo_order(index: int):
        await asyncio.sleep(index)
        return index

    async def _run():
        results = []
        for t in as_completed([echo_order(1), echo_order(2), echo_order(3)], 3):
            r = await t
            results.append(r)
        return results

    results = asyncio.run(_run())

    assert results == [1, 2, 3]



import pytest

from ragas.utils import check_if_sum_is_close, get_from_dict


@pytest.mark.parametrize(
    ["values", "close_to", "num_places"],
    [
        [[0.1, 0.2, 0.3], 0.6, 1],
        [[0.8, 0.1, 0.1], 1.0, 1],
        [[0.94, 0.03, 0.03], 1.0, 2],
        [[0.3948, 0.3948, 0.2104], 1.0, 4],
        [[10.19, 10.19, 10.19], 30.57, 2],
    ],
)
def test_check_if_sum_is_close(values, close_to, num_places):
    assert check_if_sum_is_close(values, close_to, num_places)


data_dict = {
    "something": {"nested": {"key": "value"}},
    "other": {"key": "value"},
    "key": "value",
    "another_key": "value",
    "nested_key": {"key": "value"},
}


@pytest.mark.parametrize(
    ["data_dict", "key", "expected"],
    [
        (data_dict, "something.nested.key", "value"),
        (data_dict, "other.key", "value"),
        (data_dict, "something.not_there_in_key", None),
        (data_dict, "something.nested.not_here", None),
    ],
)
def test_get_from_dict(data_dict, key, expected):
    assert get_from_dict(data_dict, key) == expected



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass, field

import numpy as np
from langchain_core.pydantic_v1 import BaseModel

from ragas.llms.output_parser import RagasoutputParser, get_json_format_instructions
from ragas.llms.prompt import Prompt, PromptValue
from ragas.metrics._answer_similarity import AnswerSimilarity
from ragas.metrics._faithfulness import (
    LONG_FORM_ANSWER_PROMPT,
    HasSegmentMethod,
    _statements_output_parser,
)
from ragas.metrics.base import (
    EvaluationMode,
    MetricWithEmbeddings,
    MetricWithLLM,
    get_segmenter,
)
from ragas.run_config import RunConfig

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks


logger = logging.getLogger(__name__)


class AnswerCorrectnessClassification(BaseModel):
    TP: t.List[t.Dict[str, t.Any]]
    FP: t.List[t.Dict[str, t.Any]]
    FN: t.List[t.Dict[str, t.Any]]


_output_instructions = get_json_format_instructions(AnswerCorrectnessClassification)
_output_parser = RagasoutputParser(pydantic_object=AnswerCorrectnessClassification)

CORRECTNESS_INSTRUCTIONS = """\
Given a ground truth and an answer statements, analyze each statement and classify them in one of the following categories:

- TP (true positive): statements that are present in answer that are also directly supported by the one or more statements in ground truth,
- FP (false positive): statements present in the answer but not directly supported by any statement in ground truth,
- FN (false negative): statements found in the ground truth but not present in answer.

Each statement can only belong to one of the categories. Provide a reason for each classification.
"""
CORRECTNESS_PROMPT = Prompt(
    name="answer_correctness",
    instruction=CORRECTNESS_INSTRUCTIONS,
    output_format_instruction=_output_instructions,
    examples=[
        {
            "question": """What powers the sun and what is its primary function?""",
            "answer": [
                "The sun is powered by nuclear fission, similar to nuclear reactors on Earth.",
                "The primary function of the sun is to provide light to the solar system.",
            ],
            "ground_truth": [
                "The sun is powered by nuclear fusion, where hydrogen atoms fuse to form helium.",
                "This fusion process in the sun's core releases a tremendous amount of energy.",
                "The energy from the sun provides heat and light, which are essential for life on Earth.",
                "The sun's light plays a critical role in Earth's climate system.",
                "Sunlight helps to drive the weather and ocean currents.",
            ],
            "classification": AnswerCorrectnessClassification.parse_obj(
                {
                    "TP": [
                        {
                            "statement": "The primary function of the sun is to provide light to the solar system.",
                            "reason": "This statement is somewhat supported by the ground truth mentioning the sun providing light and its roles, though it focuses more broadly on the sun's energy.",
                        }
                    ],
                    "FP": [
                        {
                            "statement": "The sun is powered by nuclear fission, similar to nuclear reactors on Earth.",
                            "reason": "This statement is incorrect and contradicts the ground truth which states that the sun is powered by nuclear fusion.",
                        }
                    ],
                    "FN": [
                        {
                            "statement": "The sun is powered by nuclear fusion, where hydrogen atoms fuse to form helium.",
                            "reason": "This accurate description of the sun’s power source is not included in the answer.",
                        },
                        {
                            "statement": "This fusion process in the sun's core releases a tremendous amount of energy.",
                            "reason": "This process and its significance are not mentioned in the answer.",
                        },
                        {
                            "statement": "The energy from the sun provides heat and light, which are essential for life on Earth.",
                            "reason": "The answer only mentions light, omitting the essential aspects of heat and its necessity for life, which the ground truth covers.",
                        },
                        {
                            "statement": "The sun's light plays a critical role in Earth's climate system.",
                            "reason": "This broader impact of the sun’s light on Earth's climate system is not addressed in the answer.",
                        },
                        {
                            "statement": "Sunlight helps to drive the weather and ocean currents.",
                            "reason": "The effect of sunlight on weather patterns and ocean currents is omitted in the answer.",
                        },
                    ],
                }
            ).dict(),
        },
        {
            "question": """What is the boiling point of water?""",
            "answer": [
                "The boiling point of water is 100 degrees Celsius at sea level"
            ],
            "ground_truth": [
                "The boiling point of water is 100 degrees Celsius (212 degrees Fahrenheit) at sea level.",
                "The boiling point of water can change with altitude.",
            ],
            "classification": AnswerCorrectnessClassification.parse_obj(
                {
                    "TP": [
                        {
                            "statement": "The boiling point of water is 100 degrees Celsius at sea level",
                            "reason": "This statement is directly supported by the ground truth which specifies the boiling point of water as 100 degrees Celsius at sea level.",
                        }
                    ],
                    "FP": [],
                    "FN": [
                        {
                            "statement": "The boiling point of water can change with altitude.",
                            "reason": "This additional information about how the boiling point of water can vary with altitude is not mentioned in the answer.",
                        }
                    ],
                }
            ).dict(),
        },
    ],
    input_keys=["question", "answer", "ground_truth"],
    output_key="classification",
    output_type="json",
)


@dataclass
class AnswerCorrectness(MetricWithLLM, MetricWithEmbeddings):
    """
    Measures answer correctness compared to ground truth as a combination of
    factuality and semantic similarity.

    Attributes
    ----------
    name: string
        The name of the metrics
    weights:
        a list of two weights corresponding to factuality and semantic similarity
        Defaults [0.75, 0.25]
    answer_similarity:
        The AnswerSimilarity object
    """

    name: str = "answer_correctness"  # type: ignore[reportIncompatibleMethodOverride]
    evaluation_mode: EvaluationMode = EvaluationMode.qga  # type: ignore[reportIncompatibleMethodOverride]
    correctness_prompt: Prompt = field(default_factory=lambda: CORRECTNESS_PROMPT)
    long_form_answer_prompt: Prompt = field(
        default_factory=lambda: LONG_FORM_ANSWER_PROMPT
    )
    weights: list[float] = field(default_factory=lambda: [0.75, 0.25])
    answer_similarity: t.Optional[AnswerSimilarity] = None
    sentence_segmenter: t.Optional[HasSegmentMethod] = None
    max_retries: int = 1

    def __post_init__(self: t.Self):
        if len(self.weights) != 2:
            raise ValueError(
                "Expects a list of two weights. First for factuality, second for semantic similarity"
            )
        if all([w == 0 for w in self.weights]):
            raise ValueError("At least one weight must be non-zero")
        if not all([w >= 0 for w in self.weights]):
            raise ValueError("Weights must be non-negative")

        if self.sentence_segmenter is None:
            language = self.long_form_answer_prompt.language
            self.sentence_segmenter = get_segmenter(language=language, clean=False)

    def init(self, run_config: RunConfig):
        super().init(run_config)
        if self.answer_similarity is None and self.weights[1] != 0:
            self.answer_similarity = AnswerSimilarity(
                llm=self.llm, embeddings=self.embeddings
            )

    def _compute_statement_presence(
        self, prediction: AnswerCorrectnessClassification
    ) -> float:
        tp = len(prediction.TP)
        fp = len(prediction.FP)
        fn = len(prediction.FN)
        score = tp / (tp + 0.5 * (fp + fn)) if tp > 0 else 0
        return score

    def _create_statements_prompt(self, question: str, text: str) -> PromptValue:
        assert self.sentence_segmenter is not None, "sentence_segmenter is not set"

        sentences = self.sentence_segmenter.segment(text)
        sentences = [
            sentence for sentence in sentences if sentence.strip().endswith(".")
        ]
        sentences = "\n".join([f"{i}:{x}" for i, x in enumerate(sentences)])
        prompt_value = self.long_form_answer_prompt.format(
            question=question, answer=text, sentences=sentences
        )
        return prompt_value

    async def _ascore(self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.llm is not None, "LLM must be set"

        question = row["question"]
        statements = {}
        for item in ["answer", "ground_truth"]:
            p_value = self._create_statements_prompt(question, row[item])
            item_statement = await self.llm.generate(p_value, callbacks=callbacks)
            statements[item] = await _statements_output_parser.aparse(
                item_statement.generations[0][0].text,
                p_value,
                self.llm,
                self.max_retries,
            )
            statements[item] = (
                statements[item].dicts() if statements[item] is not None else []
            )

        if not all([val == [] for val in statements.values()]):
            ground_truth = [
                statement
                for item in statements["ground_truth"]
                for statement in item["simpler_statements"]
            ]
            answer = [
                statement
                for item in statements["answer"]
                for statement in item["simpler_statements"]
            ]
            p_value = self.correctness_prompt.format(
                question=question,
                ground_truth=ground_truth,
                answer=answer,
            )
            is_statement_present = await self.llm.generate(p_value, callbacks=callbacks)
            result_text = is_statement_present.generations[0][0].text

            answers = await _output_parser.aparse(
                result_text, p_value, self.llm, self.max_retries
            )
            if answers is None:
                return np.nan

            f1_score = self._compute_statement_presence(answers)
        else:
            f1_score = 1.0

        if self.weights[1] == 0:
            similarity_score = 0.0
        else:
            assert self.answer_similarity is not None, "AnswerSimilarity must be set"

            similarity_score = await self.answer_similarity.ascore(
                row, callbacks=callbacks
            )

        score = np.average(
            [f1_score, similarity_score],
            weights=self.weights,
        )

        return float(score)

    def adapt(self, language: str, cache_dir: t.Optional[str] = None) -> None:
        assert self.llm is not None, "llm must be set to compute score"

        logger.info(f"Adapting AnswerCorrectness metric to {language}")
        self.correctness_prompt = self.correctness_prompt.adapt(
            language, self.llm, cache_dir
        )

        self.sentence_segmenter = get_segmenter(language=language, clean=False)

    def save(self, cache_dir: t.Optional[str] = None) -> None:
        self.correctness_prompt.save(cache_dir)


answer_correctness = AnswerCorrectness()



import typing as t

from langchain_core.language_models import BaseLanguageModel

from ragas.llms import llm_factory
from ragas.llms.base import BaseRagasLLM, LangchainLLMWrapper
from ragas.metrics.base import MetricWithLLM


def adapt(
    metrics: t.List[MetricWithLLM],
    language: str,
    llm: t.Optional[BaseRagasLLM] = None,
    cache_dir: t.Optional[str] = None,
) -> None:
    """
    Adapt the metric to a different language.
    """

    llm_wraper = None

    if llm is None:
        llm_wraper = llm_factory()
    elif isinstance(llm, BaseLanguageModel):
        llm_wraper = LangchainLLMWrapper(llm)
    else:
        raise ValueError("llm must be either None or a BaseLanguageModel")

    for metric in metrics:
        metric_llm = metric.llm

        if metric_llm is None or llm is not None:
            metric.llm = llm_wraper

        if hasattr(metric, "adapt"):
            metric.adapt(language, cache_dir=cache_dir)
            metric.save(cache_dir=cache_dir)
            metric.llm = metric_llm



import time

from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from llama_index.core import download_loader

from ragas.testset.evolutions import conditional, multi_context, reasoning, simple
from ragas.testset.generator import TestsetGenerator

generator_llm = ChatOpenAI(model="gpt-3.5-turbo-16k")
critic_llm = ChatOpenAI(model="gpt-4")
embeddings = OpenAIEmbeddings()

generator = TestsetGenerator.from_langchain(generator_llm, critic_llm, embeddings)

distributions = {simple: 0.5, multi_context: 0.3, reasoning: 0.1, conditional: 0.1}


def get_documents():
    SemanticScholarReader = download_loader("SemanticScholarReader")
    loader = SemanticScholarReader()
    # Narrow down the search space
    query_space = "large language models"
    # Increase the limit to obtain more documents
    documents = loader.load_data(query=query_space, limit=10)

    return documents


IGNORE_ASYNCIO = False
# os.environ["PYTHONASYNCIODEBUG"] = "1"

if __name__ == "__main__":
    documents = get_documents()

    # asyncio
    print("Starting [Asyncio]")
    start = time.time()
    generator.generate_with_llamaindex_docs(
        documents=documents,
        test_size=50,
        distributions=distributions,
        is_async=True,
    )
    print(f"Time taken: {time.time() - start:.2f}s")



import uuid

from ragas.testset.docstore import Document


def test_document_filename(monkeypatch):
    monkeypatch.setattr(uuid, "uuid4", lambda: "test-uuid")
    d1 = Document(page_content="a1")
    assert d1.filename == "test-uuid"

    # now suppose I add a filename to metadata
    d2 = Document(page_content="a2", metadata={"filename": "test-filename"})
    assert d2.filename == "test-filename"


def test_document_chunking():
    """
    Tests to make sure that there is no problem when you chunk a document into Nodes
    especially because of the fact that Node objects are created again.
    """
    from langchain.text_splitter import TokenTextSplitter
    from langchain_core.documents import Document

    from ragas.testset.docstore import Node

    splitter = TokenTextSplitter(chunk_size=1, chunk_overlap=0)
    doc = Document(page_content="Hello, world!", metadata={"filename": "test-filename"})
    nodes = [
        Node.from_langchain_document(d) for d in splitter.transform_documents([doc])
    ]
    for node in nodes:
        assert node.filename == "test-filename"



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

import numpy as np
from datasets import Dataset, concatenate_datasets
from langchain_core.callbacks import BaseCallbackHandler, BaseCallbackManager
from langchain_core.embeddings import Embeddings as LangchainEmbeddings
from langchain_core.language_models import BaseLanguageModel as LangchainLLM

from ragas._analytics import EvaluationEvent, track, track_was_completed
from ragas.callbacks import new_group
from ragas.cost import TokenUsage
from ragas.embeddings.base import (
    BaseRagasEmbeddings,
    LangchainEmbeddingsWrapper,
    embedding_factory,
)
from ragas.exceptions import ExceptionInRunner
from ragas.executor import Executor
from ragas.integrations.helicone import helicone_config
from ragas.llms import llm_factory
from ragas.llms.base import BaseRagasLLM, LangchainLLMWrapper
from ragas.metrics._answer_correctness import AnswerCorrectness
from ragas.metrics.base import (
    Metric,
    MetricWithEmbeddings,
    MetricWithLLM,
    is_reproducable,
)
from ragas.metrics.critique import AspectCritique
from ragas.run_config import RunConfig
from ragas.utils import get_feature_language, safe_nanmean

# from ragas.metrics.critique import AspectCritique
from ragas.validation import (
    handle_deprecated_ground_truths,
    remap_column_names,
    validate_column_dtypes,
    validate_evaluation_modes,
)

if t.TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks

    from ragas.cost import CostCallbackHandler, TokenUsageParser


@track_was_completed
def evaluate(
    dataset: Dataset,
    metrics: list[Metric] | None = None,
    llm: t.Optional[BaseRagasLLM | LangchainLLM] = None,
    embeddings: t.Optional[BaseRagasEmbeddings | LangchainEmbeddings] = None,
    callbacks: Callbacks = None,
    in_ci: bool = False,
    run_config: RunConfig = RunConfig(),
    token_usage_parser: t.Optional[TokenUsageParser] = None,
    raise_exceptions: bool = False,
    column_map: t.Optional[t.Dict[str, str]] = None,
) -> Result:
    """
    Run the evaluation on the dataset with different metrics

    Parameters
    ----------
    dataset : Dataset[question: list[str], contexts: list[list[str]], answer: list[str], ground_truth: list[list[str]]]
        The dataset in the format of ragas which the metrics will use to score the RAG
        pipeline with
    metrics : list[Metric] , optional
        List of metrics to use for evaluation. If not provided then ragas will run the
        evaluation on the best set of metrics to give a complete view.
    llm: BaseRagasLLM, optional
        The language model to use for the metrics. If not provided then ragas will use
        the default language model for metrics which require an LLM. This can we overridden by the llm specified in
        the metric level with `metric.llm`.
    embeddings: BaseRagasEmbeddings, optional
        The embeddings to use for the metrics. If not provided then ragas will use
        the default embeddings for metrics which require embeddings. This can we overridden by the embeddings specified in
        the metric level with `metric.embeddings`.
    callbacks: Callbacks, optional
        Lifecycle Langchain Callbacks to run during evaluation. Check the
        [langchain documentation](https://python.langchain.com/docs/modules/callbacks/)
        for more information.
    in_ci: bool
        Whether the evaluation is running in CI or not. If set to True then some
        metrics will be run to increase the reproducability of the evaluations. This
        will increase the runtime and cost of evaluations. Default is False.
    run_config: RunConfig, optional
        Configuration for runtime settings like timeout and retries. If not provided,
        default values are used.
    token_usage_parser: TokenUsageParser, optional
        Parser to get the token usage from the LLM result. If not provided then the
        the cost and total tokens will not be calculated. Default is None.
    raise_exceptions: False
        Whether to raise exceptions or not. If set to True then the evaluation will
        raise an exception if any of the metrics fail. If set to False then the
        evaluation will return `np.nan` for the row that failed. Default is False.
    column_map : dict[str, str], optional
        The column names of the dataset to use for evaluation. If the column names of
        the dataset are different from the default ones then you can provide the
        mapping as a dictionary here. Example: If the dataset column name is contexts_v1,
        column_map can be given as {"contexts":"contexts_v1"}

    Returns
    -------
    Result
        Result object containing the scores of each metric. You can use this do analysis
        later.

    Raises
    ------
    ValueError
        if validation fails because the columns required for the metrics are missing or
        if the columns are of the wrong format.

    Examples
    --------
    the basic usage is as follows:
    ```
    from ragas import evaluate

    >>> dataset
    Dataset({
        features: ['question', 'ground_truth', 'answer', 'contexts'],
        num_rows: 30
    })

    >>> result = evaluate(dataset)
    >>> print(result)
    {'context_precision': 0.817,
    'faithfulness': 0.892,
    'answer_relevancy': 0.874}
    ```
    """
    column_map = column_map or {}
    callbacks = callbacks or []

    if helicone_config.is_enabled:
        import uuid

        helicone_config.session_name = "ragas-evaluation"
        helicone_config.session_id = str(uuid.uuid4())

    if dataset is None:
        raise ValueError("Provide dataset!")

    # default metrics
    if metrics is None:
        from ragas.metrics import (
            answer_relevancy,
            context_precision,
            context_recall,
            faithfulness,
        )

        metrics = [answer_relevancy, context_precision, faithfulness, context_recall]

    # remap column names from the dataset
    dataset = remap_column_names(dataset, column_map)
    # validation
    dataset = handle_deprecated_ground_truths(dataset)
    validate_evaluation_modes(dataset, metrics)
    validate_column_dtypes(dataset)

    # set the llm and embeddings
    if isinstance(llm, LangchainLLM):
        llm = LangchainLLMWrapper(llm, run_config=run_config)
    if isinstance(embeddings, LangchainEmbeddings):
        embeddings = LangchainEmbeddingsWrapper(embeddings)

    # init llms and embeddings
    binary_metrics = []
    llm_changed: t.List[int] = []
    embeddings_changed: t.List[int] = []
    reproducable_metrics: t.List[int] = []
    answer_correctness_is_set = -1

    # loop through the metrics and perform initializations
    for i, metric in enumerate(metrics):
        # set llm and embeddings if not set
        if isinstance(metric, AspectCritique):
            binary_metrics.append(metric.name)
        if isinstance(metric, MetricWithLLM) and metric.llm is None:
            if llm is None:
                llm = llm_factory()
            metric.llm = llm
            llm_changed.append(i)
        if isinstance(metric, MetricWithEmbeddings) and metric.embeddings is None:
            if embeddings is None:
                embeddings = embedding_factory()
            metric.embeddings = embeddings
            embeddings_changed.append(i)
        if isinstance(metric, AnswerCorrectness):
            if metric.answer_similarity is None:
                answer_correctness_is_set = i
        # set reproducibility for metrics if in CI
        if in_ci and is_reproducable(metric):
            if metric.reproducibility == 1:  # type: ignore
                # only set a value if not already set
                metric.reproducibility = 3  # type: ignore
                reproducable_metrics.append(i)

        # init all the models
        metric.init(run_config)

    executor = Executor(
        desc="Evaluating",
        keep_progress_bar=True,
        raise_exceptions=raise_exceptions,
        run_config=run_config,
    )

    # Ragas Callbacks
    # init the callbacks we need for various tasks
    ragas_callbacks: t.Dict[str, BaseCallbackHandler] = {}

    # check if cost needs to be calculated
    if token_usage_parser is not None:
        from ragas.cost import CostCallbackHandler

        cost_cb = CostCallbackHandler(token_usage_parser=token_usage_parser)
        ragas_callbacks["cost_cb"] = cost_cb

    # append all the ragas_callbacks to the callbacks
    for cb in ragas_callbacks.values():
        if isinstance(callbacks, BaseCallbackManager):
            callbacks.add_handler(cb)
        else:
            callbacks.append(cb)

    # new evaluation chain
    row_run_managers = []
    evaluation_rm, evaluation_group_cm = new_group(
        name="ragas evaluation", inputs={}, callbacks=callbacks
    )
    for i, row in enumerate(dataset):
        row = t.cast(t.Dict[str, t.Any], row)
        row_rm, row_group_cm = new_group(
            name=f"row {i}",
            inputs=row,
            callbacks=evaluation_group_cm,
        )
        row_run_managers.append((row_rm, row_group_cm))
        [
            executor.submit(
                metric.ascore,
                row,
                row_group_cm,
                name=f"{metric.name}-{i}",
                timeout=run_config.timeout,
            )
            for metric in metrics
        ]

    scores = []
    try:
        # get the results
        results = executor.results()
        if results == []:
            raise ExceptionInRunner()

        # convert results to dataset_like
        for i, _ in enumerate(dataset):
            s = {}
            for j, m in enumerate(metrics):
                s[m.name] = results[len(metrics) * i + j]
            scores.append(s)
            # close the row chain
            row_rm, row_group_cm = row_run_managers[i]
            if not row_group_cm.ended:
                row_rm.on_chain_end(s)

    # run evaluation task
    except Exception as e:
        if not evaluation_group_cm.ended:
            evaluation_rm.on_chain_error(e)

        raise e
    else:
        # evalution run was successful
        # now lets process the results
        cost_cb = ragas_callbacks["cost_cb"] if "cost_cb" in ragas_callbacks else None
        result = Result(
            scores=Dataset.from_list(scores),
            dataset=dataset,
            binary_columns=binary_metrics,
            cost_cb=t.cast(
                t.Union["CostCallbackHandler", None],
                cost_cb,
            ),
        )
        if not evaluation_group_cm.ended:
            evaluation_rm.on_chain_end(result)
    finally:
        # reset llms and embeddings if changed
        for i in llm_changed:
            t.cast(MetricWithLLM, metrics[i]).llm = None
        for i in embeddings_changed:
            t.cast(MetricWithEmbeddings, metrics[i]).embeddings = None
        if answer_correctness_is_set != -1:
            t.cast(
                AnswerCorrectness, metrics[answer_correctness_is_set]
            ).answer_similarity = None

        for i in reproducable_metrics:
            metrics[i].reproducibility = 1  # type: ignore

    # log the evaluation event
    metrics_names = [m.name for m in metrics]
    metric_lang = [get_feature_language(m) for m in metrics]
    metric_lang = np.unique([m for m in metric_lang if m is not None])
    track(
        EvaluationEvent(
            event_type="evaluation",
            metrics=metrics_names,
            evaluation_mode="",
            num_rows=dataset.shape[0],
            language=metric_lang[0] if len(metric_lang) > 0 else "",
            in_ci=in_ci,
        )
    )
    return result


@dataclass
class Result(dict):
    scores: Dataset
    dataset: t.Optional[Dataset] = None
    binary_columns: t.List[str] = field(default_factory=list)
    cost_cb: t.Optional[CostCallbackHandler] = None

    def __post_init__(self):
        values = []
        for cn in self.scores[0].keys():
            value = safe_nanmean(self.scores[cn])
            self[cn] = value
            if cn not in self.binary_columns:
                value = t.cast(float, value)
                values.append(value + 1e-10)

    def to_pandas(self, batch_size: int | None = None, batched: bool = False):
        if self.dataset is None:
            raise ValueError("dataset is not provided for the results class")
        assert self.scores.shape[0] == self.dataset.shape[0]
        result_ds = concatenate_datasets([self.dataset, self.scores], axis=1)

        return result_ds.to_pandas(batch_size=batch_size, batched=batched)

    def total_tokens(self) -> t.Union[t.List[TokenUsage], TokenUsage]:
        if self.cost_cb is None:
            raise ValueError(
                "The evaluate() run was not configured for computing cost. Please provide a token_usage_parser function to evaluate() to compute cost."
            )
        return self.cost_cb.total_tokens()

    def total_cost(
        self,
        cost_per_input_token: t.Optional[float] = None,
        cost_per_output_token: t.Optional[float] = None,
        per_model_costs: t.Dict[str, t.Tuple[float, float]] = {},
    ) -> float:
        if self.cost_cb is None:
            raise ValueError(
                "The evaluate() run was not configured for computing cost. Please provide a token_usage_parser function to evaluate() to compute cost."
            )
        return self.cost_cb.total_cost(
            cost_per_input_token, cost_per_output_token, per_model_costs
        )

    def __repr__(self) -> str:
        scores = self.copy()
        score_strs = [f"'{k}': {v:0.4f}" for k, v in scores.items()]
        return "{" + ", ".join(score_strs) + "}"



from uuid import UUID

import pandas as pd
import pytest
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.runnables import (
    Runnable,
    RunnableLambda,
    RunnableParallel,
    RunnablePassthrough,
)
from langchain_openai import ChatOpenAI
from langsmith import Client

client = Client()


# datarows with queries and ground truth
# sample questions
questions = [
    "What are the advantages of remote work? Why does gitlab do it?",
    "what are the dis-advantages of remote work in gitlab? How do you mitigate that?",
    "What does it mean to be 'all-remote'? Why is it important?",
    "How does compensation work in an 'all-remote' setting?",
    "How to run effective meetings in 'all-remote' setting",
]

answers = [
    "Remote work offers numerous advantages including increased flexibility, the ability to hire top talent globally without geographical constraints, enhanced productivity due to fewer distractions, and significant cost savings on office space and related expenses. GitLab adopts an all-remote model to leverage these benefits, ensuring a more inclusive and diverse workforce, fostering a culture that values output over presence, and providing employees the freedom to work in environments that best suit their personal and professional needs. This approach not only supports individual employee well-being and work-life balance but also positions GitLab as a resilient and adaptive organization in a rapidly changing global work landscape.",
    "GitLab's remote work challenges include onboarding difficulties, loneliness, communication breakdowns, work/life balance issues, time zone conflicts, and the need for strong time management skills. To mitigate these, GitLab employs strategies such as providing comprehensive onboarding resources, fostering community through virtual coffee chats and social calls, prioritizing asynchronous communication, reimbursing coworking spaces, empowering employees to manage their own schedules, focusing on results rather than hours, and screening for self-discipline during the hiring process. These measures aim to enhance productivity and employee satisfaction in a remote work setting.",
    "Being 'all-remote' means that an organization empowers every individual to work from any location where they feel most fulfilled, without the need to report to a company-owned office, thereby treating all employees equally regardless of their physical location. This approach is important because it eliminates location hierarchy, allowing for a more inclusive work environment where team members have the autonomy to create their ideal workspace and can maintain their job regardless of life changes, such as relocations due to family commitments. It supports a diverse workforce, including caregivers, working parents, and military spouses, by providing them with the flexibility to work from anywhere, fostering equality among all employees and enabling a global talent pool without the constraints of geographical boundaries.",
    "In an 'all-remote' setting, such as at GitLab, compensation is structured around local rates rather than a single global standard, which means employees are paid based on the cost of living and market rates in their respective locations. This approach allows the company to hire globally without being bound by the high salary standards of any particular region, like San Francisco. GitLab uses a compensation calculator to ensure transparency and fairness in pay, adjusting salaries based on a combination of factors including location, experience, and market data. Payments are typically made in the local currency of the employee, and for countries where direct employment isn't feasible, GitLab utilizes professional employment organizations or hires contractors. This model supports GitLab's global talent acquisition strategy while managing compensation costs effectively.",
    "To run effective meetings in an 'all-remote' setting, it's crucial to be intentional about meeting necessity, provide clear agendas and supporting materials in advance, start and end on time, document discussions in real time, and make attendance optional to respect time zones and individual schedules. Recording meetings for asynchronous viewing, using reliable communication tools like Zoom, and ensuring active participation through video feedback are also key practices. This approach aligns with GitLab's guidelines for maximizing efficiency, inclusivity, and collaboration in a remote work environment.",
]

dataset = {"question": questions, "ground_truth": answers}


def upload_to_langsmith(dataset_name: str) -> UUID:
    # Creating a pandas DataFrame from the dataset dictionary
    df = pd.DataFrame(dataset)

    # upload to langsmith
    langsmith_dataset = client.upload_dataframe(
        name=dataset_name,
        description="temporal dataset for testing langsmith",
        df=df,
        input_keys=["question"],
        output_keys=["ground_truth"],
    )

    return langsmith_dataset.id


def clean_langsmith(langsmith_dataset_id: UUID):
    # clean langsmith
    client.delete_dataset(dataset_id=langsmith_dataset_id)


def llm_chain_factory() -> Runnable:
    # just LLM
    template = """Use the following pieces of context to answer the question at the end.
    If you don't know the answer, just say that you don't know, don't try to make up an answer.
    Use three sentences maximum and keep the answer as concise as possible.
    Always say "thanks for asking!" at the end of the answer.

    Question: {question}

    Helpful Answer:"""
    llm_prompt = PromptTemplate.from_template(template)

    # LLM
    llm = ChatOpenAI(model="gpt-3.5-turbo", temperature=0)

    # just llm pipeline
    just_llm = (
        {"question": RunnablePassthrough()}
        | llm_prompt
        | llm
        | StrOutputParser()
        | RunnableParallel(
            {
                "answer": RunnablePassthrough(),
                "contexts": RunnableLambda(lambda _: [""]),
            }
        )
    )

    return just_llm


@pytest.fixture()
def langsmith_dataset():
    dataset_name = "temporal_dataset"
    langsmith_dataset_id = upload_to_langsmith(dataset_name)
    yield dataset_name
    clean_langsmith(langsmith_dataset_id)


@pytest.mark.e2e()
def test_langsmith_evaluate(langsmith_dataset):
    # setup
    just_llm = llm_chain_factory()

    from ragas.integrations.langsmith import evaluate
    from ragas.metrics import answer_correctness

    # evaluate just llms
    _ = evaluate(
        dataset_name=langsmith_dataset,
        llm_or_chain_factory=just_llm,
        # experiment_name="just_llm",
        metrics=[answer_correctness],
        verbose=True,
    )



from __future__ import annotations

import ast
import json
import logging
import os
import typing as t

from langchain_core.messages import BaseMessage, HumanMessage
from langchain_core.prompt_values import PromptValue as BasePromptValue
from langchain_core.pydantic_v1 import BaseModel, root_validator

from ragas.llms import BaseRagasLLM
from ragas.llms.json_load import json_loader
from ragas.utils import get_cache_dir

Example = t.Dict[str, t.Any]

logger = logging.getLogger(__name__)


class PromptValue(BasePromptValue):
    prompt_str: str

    def to_messages(self) -> t.List[BaseMessage]:
        """Return prompt as a list of Messages."""
        return [HumanMessage(content=self.to_string())]

    def to_string(self) -> str:
        return self.prompt_str


class Prompt(BaseModel):
    """
    Prompt is a class that represents a prompt for the ragas metrics.

    Attributes:
        name (str): The name of the prompt.
        instruction (str): The instruction for the prompt.
        output_format_instruction (str): The output format instruction for the prompt.
        examples (List[Dict[str, Any]]): List of example inputs and outputs for the prompt.
        input_keys (List[str]): List of input variable names.
        output_key (str): The output variable name.
        output_type (Literal["json", "str"]): The type of the output (default: "json").
        language (str): The language of the prompt (default: "english").
    """

    name: str = ""
    instruction: str
    output_format_instruction: str = ""
    examples: t.List[Example] = []
    input_keys: t.List[str] = [""]
    output_key: str = ""
    output_type: t.Literal["json", "str"] = "json"
    language: str = "english"

    @root_validator
    def validate_prompt(cls, values: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        """
        Validate the template string to ensure that it is in desired format.
        """
        if values.get("instruction") is None or values.get("instruction") == "":
            raise ValueError("instruction cannot be empty")
        if values.get("input_keys") is None or values.get("instruction") == []:
            raise ValueError("input_keys cannot be empty")
        if values.get("output_key") is None or values.get("output_key") == "":
            raise ValueError("output_key cannot be empty")

        if values.get("examples"):
            output_key = values["output_key"]
            for no, example in enumerate(values["examples"]):
                for inp_key in values["input_keys"]:
                    if inp_key not in example:
                        raise ValueError(
                            f"example {no+1} does not have the variable {inp_key} in the definition"
                        )
                if output_key not in example:
                    raise ValueError(
                        f"example {no+1} does not have the variable {output_key} in the definition"
                    )
                if values["output_type"].lower() == "json":
                    try:
                        if output_key in example:
                            if isinstance(example[output_key], str):
                                json.loads(example[output_key])
                    except ValueError as e:
                        raise ValueError(
                            f"{output_key} in example {no+1} is not in valid json format: {e}"
                        )

        return values

    def to_string(self) -> str:
        """
        Generate the prompt string from the variables.
        """
        prompt_elements = [self.instruction]
        if self.output_format_instruction:
            prompt_elements.append(
                "\n"
                + self.output_format_instruction.replace("{", "{{").replace("}", "}}")
            )
        prompt_str = "\n".join(prompt_elements) + "\n"

        if self.examples:
            prompt_str += "\nExamples:\n"
            # Format the examples to match the Langchain prompt template
            for example in self.examples:
                for key, value in example.items():
                    is_json = isinstance(value, (dict, list))
                    value = (
                        json.dumps(value, ensure_ascii=False).encode("utf8").decode()
                    )
                    value = (
                        value.replace("{", "{{").replace("}", "}}")
                        if self.output_type.lower() == "json"
                        else value
                    )
                    prompt_str += (
                        f"\n{key}: {value}"
                        if not is_json
                        else f"\n{key}: ```{value}```"
                    )
                prompt_str += "\n"

        prompt_str += "\nYour actual task:\n"

        if self.input_keys:
            prompt_str += "".join(f"\n{key}: {{{key}}}" for key in self.input_keys)
        if self.output_key:
            prompt_str += f"\n{self.output_key}: \n"

        return prompt_str

    def get_example_str(self, example_no: int) -> str:
        """
        Get the example string from the example number.
        """
        if example_no >= len(self.examples):
            raise ValueError(f"example number {example_no} is out of range")
        example = self.examples[example_no]
        example_str = ""
        for key, value in example.items():
            value = json.dumps(value, ensure_ascii=False).encode("utf8").decode()
            value = (
                value.replace("{", "{{").replace("}", "}}")
                if self.output_type.lower() == "json"
                else value
            )
            example_str += f"\n{key}: {value}"
        return "```" + example_str + "```"

    def format(self, **kwargs: t.Any) -> PromptValue:
        """
        Format the Prompt object into a ChatPromptTemplate object to be used in metrics.
        """
        if set(self.input_keys) != set(kwargs.keys()):
            raise ValueError(
                f"Input variables {self.input_keys} do not match with the given parameters {list(kwargs.keys())}"
            )
        for key, value in kwargs.items():
            if isinstance(value, str):
                kwargs[key] = json.dumps(value)

        prompt = self.to_string()
        return PromptValue(prompt_str=prompt.format(**kwargs))

    def adapt(
        self, language: str, llm: BaseRagasLLM, cache_dir: t.Optional[str] = None
    ) -> Prompt:
        def get_all_keys(nested_json):
            keys = set()
            for key, value in nested_json.items():
                keys.add(key)
                if isinstance(value, dict):
                    keys = keys.union(get_all_keys(value))
            return keys

        if self.language == language:
            return self

        # TODO: Add callbacks
        cache_dir = cache_dir if cache_dir else get_cache_dir()
        if os.path.exists(os.path.join(cache_dir, language, f"{self.name}.json")):
            self_cp = self._load(language, self.name, cache_dir)

            self.language = self_cp.language
            self.examples = self_cp.examples

            return self_cp

        logger.info("Adapting %s to %s", self.name, language)
        prompts = []
        output_keys = []
        for example in self.examples:
            prompts.extend(
                [
                    str_translation.format(
                        translate_to=language, input=example.get(key)
                    )
                    for key in self.input_keys
                ]
            )
            prompts.append(
                json_translatation.format(
                    translate_to=language, input=example.get(self.output_key)
                )
                if self.output_type.lower() == "json"
                else str_translation.format(
                    translate_to=language, input=example.get(self.output_key)
                )
            )
            if self.output_type.lower() == "json":
                output = example.get(self.output_key)
                if isinstance(output, str):
                    output = json.loads(output)
                if isinstance(output, dict):
                    output_keys.append(get_all_keys(output))
                elif isinstance(output, list) and all(
                    isinstance(item, dict) for item in output
                ):
                    output_keys.append([get_all_keys(item) for item in output])

        # NOTE: this is a slow loop, consider Executor to fasten this
        results = []
        for p in prompts:
            results.append(llm.generate_text(p).generations[0][0].text)
        per_example_items = len(self.input_keys) + 1
        grouped_results = [
            results[i : i + per_example_items]
            for i in range(0, len(results), per_example_items)
        ]
        assert len(grouped_results) == len(
            self.examples
        ), "examples and adapted examples must be of equal length"
        for i, example in enumerate(grouped_results):
            example_dict = {}
            example_dict.update(
                {k: v for k, v in zip(self.input_keys, example[: len(self.input_keys)])}
            )
            if self.output_type.lower() == "json":
                example_dict[self.output_key] = json_loader._safe_load(example[-1], llm)
                if example_dict[self.output_key] == {}:
                    # Extracting the dictionary part using string slicing
                    dict_str = example[-1].split("(")[0].strip()
                    example_dict[self.output_key] = ast.literal_eval(dict_str)
                else:
                    example_dict[self.output_key] = example[-1]
            if self.output_type.lower() == "json":
                output = example_dict[self.output_key]
                if isinstance(output, dict):
                    assert (
                        set(output.keys()) == output_keys[i]
                    ), f"Adapted output keys {set(output.keys())=} do not match with the original output keys: {output_keys[i]=}"
                elif isinstance(output, list) and all(
                    isinstance(item, dict) for item in output
                ):
                    assert all(
                        set(item.keys()) in output_keys[i] for item in output
                    ), "Adapted output keys do not match with the original output keys"

            self.examples[i] = example_dict

        self.language = language

        # TODO:Validate the prompt after adaptation

        self.save(cache_dir=cache_dir)

        return self

    def save(self, cache_dir: t.Optional[str] = None):
        cache_dir = cache_dir if cache_dir else get_cache_dir()
        cache_dir = os.path.join(cache_dir, self.language)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        cache_path = os.path.join(cache_dir, f"{self.name}.json")
        with open(cache_path, "w") as file:
            json.dump(self.dict(), file, indent=4)

    @classmethod
    def _load(cls, language: str, name: str, cache_dir: str) -> Prompt:
        logger.info("Loading %s from %s", name, cache_dir)
        path = os.path.join(cache_dir, language, f"{name}.json")
        return cls(**json.load(open(path)))


str_translation = Prompt(
    name="str_translation",
    instruction="Language translation",
    examples=[
        {
            "translate_to": "hindi",
            "input": "Who was  Albert Einstein and what is he best known for?",
            "output": "अल्बर्ट आइंस्टीन कौन थे और वे किसके लिए सबसे ज्यादा प्रसिद्ध हैं?",
        },
        {
            "translate_to": "dutch",
            "input": "Who was queen Elizabeth and what is she best known for?",
            "output": "Wie was koningin Elizabeth en waar is zij het meest bekend om?",
        },
    ],
    input_keys=["translate_to", "input"],
    output_key="output",
    output_type="str",
)

json_translatation = Prompt(
    name="json_translation",
    instruction="Translate values in given json to target language and output the translated json",
    examples=[
        {
            "translate_to": "hindi",
            "input": {
                "statements": [
                    "Albert Einstein was born in Germany.",
                    "Albert Einstein was best known for his theory of relativity.",
                ]
            },
            "output": {
                "statements": [
                    "अल्बर्ट आइंस्टीन का जन्म जर्मनी में हुआ था।",
                    "अल्बर्ट आइंस्टीन अपने सापेक्षता के सिद्धांत के लिए सबसे अधिक प्रसिद्ध थे।",
                ]
            },
        },
        {
            "translate_to": "dutch",
            "input": {
                "statements": [
                    "Paris is the capital of France.",
                    "Croissants are a popular French pastry.",
                ]
            },
            "output": {
                "statements": [
                    "Parijs is de hoofdstad van Frankrijk.",
                    "Croissants zijn een populair Frans gebak.",
                ]
            },
        },
    ],
    input_keys=["translate_to", "input"],
    output_key="output",
    output_type="json",
)



import typing as t

from langchain_core.documents import Document as LCDocument
from ragas_experimental.testset.extractors.base import Extractor
from ragas_experimental.testset.extractors.llm_based import LLMbasedExtractor
from ragas_experimental.testset.extractors.regex_based import RulebasedExtractor
from ragas_experimental.testset.graph import Node, NodeLevel

from ragas.embeddings.base import BaseRagasEmbeddings, embedding_factory
from ragas.executor import Executor
from ragas.llms.base import BaseRagasLLM, llm_factory


class DocumentExtractor:
    def __init__(
        self,
        extractors: t.List[Extractor],
        llm: t.Optional[BaseRagasLLM] = None,
        embedding: t.Optional[BaseRagasEmbeddings] = None,
    ):
        self.extractors = extractors
        self.llm = llm or llm_factory()
        self.embedding = embedding or embedding_factory()

    @property
    def extractors(self):
        return self._extractors

    @extractors.setter
    def extractors(self, extractors):
        self._extractors = []
        llm_extractors = [
            extractor
            for extractor in extractors
            if isinstance(extractor, LLMbasedExtractor)
        ]
        rule_extractor = [
            extractor
            for extractor in extractors
            if isinstance(extractor, RulebasedExtractor)
        ]
        if llm_extractors:
            self._extractors.extend(LLMbasedExtractor.merge_extractors(*llm_extractors))
        if rule_extractor:
            self._extractors.extend(
                RulebasedExtractor.merge_extractors(*rule_extractor)
            )

    def extract(
        self, inputs: t.Sequence[Node | LCDocument], *args
    ) -> t.Sequence[Node | LCDocument]:
        exec = Executor(
            desc="Extraction..",
            keep_progress_bar=True,
            raise_exceptions=True,
            run_config=None,
        )

        if all(isinstance(item, Node) for item in inputs):
            for item in inputs:
                exec.submit(self._extract_from_node, item, *args)
        elif all(isinstance(item, LCDocument) for item in inputs):
            for item in inputs:
                exec.submit(self._extract_from_document, item)
        else:
            raise ValueError("Input must be a list of Nodes or Documents")

        return exec.results()

    def embed(
        self,
        inputs: t.Sequence[Node | LCDocument],
        attributes: t.List[str],
        args: t.Dict[str, t.List[NodeLevel]] = {},
    ) -> t.Sequence[Node | LCDocument]:
        exec = Executor(
            desc="Embeddding..",
            keep_progress_bar=True,
            raise_exceptions=True,
            run_config=None,
        )

        for attr in attributes:
            if all(isinstance(item, Node) for item in inputs):
                for item in inputs:
                    exec.submit(
                        self._embed_from_node, item, [attr], args.get(attr, "any")
                    )
            elif all(isinstance(item, LCDocument) for item in inputs):
                for item in inputs:
                    exec.submit(self._embed_document, item, [attr])
            else:
                raise ValueError("Input must be a list of Nodes or Documents")

        return exec.results()

    async def _extract_from_document(self, doc: LCDocument) -> LCDocument:
        for extractor in self._extractors:
            if isinstance(extractor, LLMbasedExtractor):
                output = await extractor.aextract(doc)
                doc.metadata.update(output)
            elif isinstance(extractor, RulebasedExtractor):
                output = extractor.extract(doc)
                doc.metadata.update(output)
            else:
                raise ValueError("Extractor not supported")

        extractive_metadata_keys = []
        for metadata in doc.metadata:
            if isinstance(doc.metadata[metadata], str):
                idx = doc.page_content.find(doc.metadata[metadata])
                if idx != -1:
                    extractive_metadata_keys.append(metadata)
            elif isinstance(doc.metadata[metadata], list) and all(
                isinstance(item, str) for item in doc.metadata[metadata]
            ):
                idx = [doc.page_content.find(item) for item in doc.metadata[metadata]]
                if sum(i != -1 for i in idx) > len(idx) / 2:
                    extractive_metadata_keys.append(metadata)

        doc.metadata["extractive_metadata_keys"] = extractive_metadata_keys

        return doc

    async def _extract_from_node(
        self, node: Node, levels: t.Union[str, t.List[NodeLevel]] = "any"
    ) -> Node:
        if levels == "any" or (isinstance(levels, list) and node.level in levels):
            for extractor in self._extractors:
                if isinstance(extractor, LLMbasedExtractor):
                    output = await extractor.aextract(node)
                    node.properties["metadata"].update(output)
                elif isinstance(extractor, RulebasedExtractor):
                    output = extractor.extract(node)
                    node.properties["metadata"].update(output)
                else:
                    raise ValueError("Extractor not supported")

        return node

    async def _embed_document(
        self, doc: LCDocument, attributes: t.List[str]
    ) -> LCDocument:
        self.embedding = (
            self.embedding if self.embedding is not None else embedding_factory()
        )
        for attr in attributes:
            if attr == "page_content":
                item_to_embed = doc.page_content
            elif attr in doc.metadata:
                item_to_embed = doc.metadata.get(attr, "")
            else:
                raise ValueError(f"Attribute {attr} not found in document")

            embedding = await self.embedding.aembed_query(item_to_embed)
            doc.metadata[f"{attr}_embedding"] = embedding

        return doc

    async def _embed_from_node(
        self,
        node: Node,
        attributes: t.List[str],
        levels: t.Union[str, t.List[NodeLevel]] = "any",
    ) -> Node:
        if levels == "any" or (isinstance(levels, list) and node.level in levels):
            self.embedding = (
                self.embedding if self.embedding is not None else embedding_factory()
            )
            for attr in attributes:
                if attr == "page_content":
                    item_to_embed = node.properties["page_content"]
                elif attr in node.properties["metadata"]:
                    item_to_embed = node.properties["metadata"][attr]
                else:
                    raise ValueError(f"Attribute {attr} not found in document")

                embedding = await self.embedding.aembed_query(item_to_embed)
                node.properties["metadata"][f"{attr}_embedding"] = embedding

        return node



from datasets import Dataset

from ragas.metrics._answer_correctness import answer_correctness
from ragas.metrics._answer_relevance import answer_relevancy
from ragas.metrics._answer_similarity import answer_similarity
from ragas.metrics._context_entities_recall import context_entity_recall
from ragas.metrics._context_precision import context_precision, context_utilization
from ragas.metrics._context_recall import context_recall
from ragas.metrics._faithfulness import faithfulness
from ragas.metrics.base import Metric
from ragas.validation import validate_evaluation_modes

ALL_METRICS = [
    answer_correctness,
    faithfulness,
    answer_similarity,
    context_precision,
    context_utilization,
    context_recall,
    answer_relevancy,
    context_entity_recall,
]


def get_available_metrics(ds: Dataset) -> list[Metric]:
    """
    Get the available metrics for the given dataset.
    E.g. if the dataset contains ("question", "answer", "contexts") columns,
    the available metrics are those that can be evaluated in [qa, qac, qc] mode.
    """
    available_metrics = []
    for metric in ALL_METRICS:
        try:
            validate_evaluation_modes(ds, [metric])
            available_metrics.append(metric)
        except ValueError:
            continue
    return available_metrics



import json
import logging
import typing as t

from langchain_core.exceptions import OutputParserException
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.pydantic_v1 import BaseModel

from ragas.llms import BaseRagasLLM
from ragas.llms.prompt import Prompt, PromptValue

logger = logging.getLogger(__name__)
# The get_format_instructions function is a modified version from
# langchain_core.output_parser.pydantic. The original version removed the "type" json schema
# property that confused some older LLMs.

TBaseModel = t.TypeVar("TBaseModel", bound=BaseModel)

FIX_OUTPUT_FORMAT = Prompt(
    name="",
    instruction="Below, the Completion did not satisfy the constraints given in the Prompt.",
    output_format_instruction="",
    input_keys=["prompt", "completion"],
    output_key="fixed_completion",
)


JSON_FORMAT_INSTRUCTIONS = """The output should be a well-formatted JSON instance that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output JSON schema:
```
{schema}
```

Do not return any preamble or explanations, return only a pure JSON string surrounded by triple backticks (```)."""


def get_json_format_instructions(pydantic_object: t.Type[TBaseModel]) -> str:
    # Copy schema to avoid altering original Pydantic schema.
    schema = {k: v for k, v in pydantic_object.schema().items()}

    # Remove extraneous fields.
    reduced_schema = schema
    if "title" in reduced_schema:
        del reduced_schema["title"]
    # Ensure json in context is well-formed with double quotes.
    schema_str = json.dumps(reduced_schema)

    resp = JSON_FORMAT_INSTRUCTIONS.format(schema=schema_str)
    return resp


class RagasoutputParser(PydanticOutputParser):
    async def aparse(  # type: ignore
        self, result: str, prompt: PromptValue, llm: BaseRagasLLM, max_retries: int = 1
    ):
        try:
            output = super().parse(result)
        except OutputParserException:
            if max_retries != 0:
                p_value = FIX_OUTPUT_FORMAT.format(
                    prompt=prompt.to_string(), completion=result
                )
                output = await llm.generate(p_value)
                result = output.generations[0][0].text
                return await self.aparse(result, prompt, llm, max_retries - 1)
            else:
                logger.warning("Failed to parse output. Returning None.")
                return None
        return output



import re
import typing as t
from collections import defaultdict
from dataclasses import dataclass

from langchain_core.documents import Document as LCDocument
from ragas_experimental.testset.extractors.base import Extractor, Regex
from ragas_experimental.testset.graph import Node


@dataclass
class RulebasedExtractor(Extractor):
    regex: t.Optional[Regex] = None
    is_multiline: bool = False

    def __post_init__(self):
        assert self.regex is not None, "Regex pattern is not initialized"
        self.pattern = self.regex()

    async def aextract_text(self, text):
        raise NotImplementedError(
            "aextract() is not implemented for RulebasedExtractor"
        )

    def extract_text(self, text):
        matches = (
            re.finditer(self.pattern, text, re.MULTILINE)
            if self.is_multiline
            else re.finditer(self.pattern, text)
        )
        result = defaultdict(list)
        for m in matches:
            m_dict = {k: v for k, v in m.groupdict().items() if v is not None}
            for key, value in m_dict.items():
                result[key].append(value)
        return result

    def extract(self, node: t.Union[Node, LCDocument]) -> t.Any:
        return super().extract(node)

    def merge_extractors(self, *extractors) -> t.List[Extractor]:
        if isinstance(self, RulebasedExtractor):
            extractors = (self,) + extractors

        assert all(
            isinstance(extractor, RulebasedExtractor) for extractor in extractors
        ), "All extractors must be of type RulebasedExtractor"

        final_extractors: t.List[t.List[RulebasedExtractor]] = []
        added_indices = []
        for idx, extractor in enumerate(extractors):
            if idx not in added_indices:
                final_extractors.append([extractor])
                added_indices.append(idx)
                other_extractors = [
                    ext for i, ext in enumerate(extractors) if i not in added_indices
                ]
                filtered_extractors = [
                    ext
                    for ext in other_extractors
                    if extractor.attribute == ext.attribute
                    if extractor.is_multiline == ext.is_multiline
                ]
                for ext in filtered_extractors:
                    final_extractors[-1].append(ext)
                    added_indices.append(extractors.index(ext))

        extractors_to_return = []
        for group_index, extractors in enumerate(final_extractors):
            if len(extractors) > 1:
                # Process each pattern individually
                processed_patterns = []
                for extractor in extractors:
                    pattern = extractor.pattern
                    # Extract flags from the beginning of the pattern
                    flags = ""
                    if pattern.startswith("(?"):
                        flag_end = pattern.index(")")
                        flags = pattern[2:flag_end]
                        pattern = pattern[flag_end + 1 :]
                    # Wrap the pattern in a non-capturing group with flags
                    processed_patterns.append(f"(?{flags}:{pattern})")

                # Join all processed patterns
                merged_pattern = "|".join(processed_patterns)

                updated_regex = Regex(name="merged_extractor", pattern=merged_pattern)
            else:
                updated_regex = extractors[0].regex

            extractors_to_return.append(
                RulebasedExtractor(
                    attribute=extractors[0].attribute,
                    regex=updated_regex,
                    is_multiline=extractors[0].is_multiline,
                )
            )
        return extractors_to_return


links_extractor_pattern = r"(?i)\b(?:https?://|www\.)\S+\b"
emails_extractor_pattern = r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+"
markdown_headings_pattern = r"^(#{1,6})\s+(.*)"

email_extractor = RulebasedExtractor(
    regex=Regex(name="email", pattern=emails_extractor_pattern)
)
link_extractor = RulebasedExtractor(
    regex=Regex(name="link", pattern=links_extractor_pattern)
)
markdown_headings = RulebasedExtractor(
    regex=Regex(name="markdown_headings", pattern=markdown_headings_pattern),
    is_multiline=True,
)



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass

import numpy as np

from ragas.embeddings.base import HuggingfaceEmbeddings
from ragas.metrics.base import EvaluationMode, MetricWithEmbeddings, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain_core.callbacks.base import Callbacks


logger = logging.getLogger(__name__)


@dataclass
class AnswerSimilarity(MetricWithLLM, MetricWithEmbeddings):
    """
    Scores the semantic similarity of ground truth with generated answer.
    cross encoder score is used to quantify semantic similarity.
    SAS paper: https://arxiv.org/pdf/2108.06130.pdf

    Attributes
    ----------
    name : str
    model_name:
        The model to be used for calculating semantic similarity
        Defaults open-ai-embeddings
        select cross-encoder model for best results
        https://huggingface.co/spaces/mteb/leaderboard
    threshold:
        The threshold if given used to map output to binary
        Default 0.5
    """

    name: str = "answer_similarity"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.ga  # type: ignore
    is_cross_encoder: bool = False
    threshold: t.Optional[float] = None

    def __post_init__(self: t.Self):
        # only for cross encoder
        if isinstance(self.embeddings, HuggingfaceEmbeddings):
            self.is_cross_encoder = True if self.embeddings.is_cross_encoder else False
            self.embeddings.encode_kwargs = {
                **self.embeddings.encode_kwargs,
            }

    async def _ascore(self: t.Self, row: t.Dict, callbacks: Callbacks) -> float:
        assert self.embeddings is not None, "embeddings must be set"

        ground_truth = t.cast(str, row["ground_truth"])
        answer = t.cast(str, row["answer"])

        if self.is_cross_encoder and isinstance(self.embeddings, HuggingfaceEmbeddings):
            raise NotImplementedError(
                "async score [ascore()] not implemented for HuggingFace embeddings"
            )
        else:
            embedding_1 = np.array(await self.embeddings.embed_text(ground_truth))
            embedding_2 = np.array(await self.embeddings.embed_text(answer))
            # Normalization factors of the above embeddings
            norms_1 = np.linalg.norm(embedding_1, keepdims=True)
            norms_2 = np.linalg.norm(embedding_2, keepdims=True)
            embedding_1_normalized = embedding_1 / norms_1
            embedding_2_normalized = embedding_2 / norms_2
            similarity = embedding_1_normalized @ embedding_2_normalized.T
            score = similarity.flatten()

        assert isinstance(score, np.ndarray), "Expects ndarray"
        if self.threshold:
            score = score >= self.threshold

        return score.tolist()[0]


answer_similarity = AnswerSimilarity()



"""Async utils."""

import asyncio
from typing import Any, Coroutine, List


def run_async_tasks(
    tasks: List[Coroutine],
    show_progress: bool = False,
    progress_bar_desc: str = "Running async tasks",
) -> List[Any]:
    """Run a list of async tasks."""
    tasks_to_execute: List[Any] = tasks

    # if running in notebook, use nest_asyncio to hijack the event loop
    try:
        loop = asyncio.get_running_loop()
        try:
            import nest_asyncio
        except ImportError:
            raise RuntimeError(
                "nest_asyncio is required to run async tasks in jupyter. Please install it via `pip install nest_asyncio`."  # noqa
            )
        else:
            nest_asyncio.apply()
    except RuntimeError:
        loop = asyncio.new_event_loop()

    # gather tasks to run
    if show_progress:
        from tqdm.asyncio import tqdm

        async def _gather() -> List[Any]:
            "gather tasks and show progress bar"
            return await tqdm.gather(*tasks_to_execute, desc=progress_bar_desc)

    else:  # don't show_progress

        async def _gather() -> List[Any]:
            return await asyncio.gather(*tasks_to_execute)

    try:
        outputs: List[Any] = loop.run_until_complete(_gather())
    except Exception as e:
        # run the operation w/o tqdm on hitting a fatal
        # may occur in some environments where tqdm.asyncio
        # is not supported
        raise RuntimeError("Fatal error occurred while running async tasks.", e) from e
    return outputs



import typing as t
import uuid

from graphene import (
    Argument,
    Enum,
    Field,
    InputObjectType,
    JSONString,
    List,
    ObjectType,
    Schema,
    String,
)


class NodeType(Enum):
    DOC = "doc"
    CHUNK = "chunk"


class NodeLevel(Enum):
    LEVEL_0 = 0
    LEVEL_1 = 1
    LEVEL_2 = 2
    LEVEL_3 = 3
    LEVEL_4 = 4
    LEVEL_5 = 5

    def next_level(self) -> t.Optional["NodeLevel"]:
        level_values = list(NodeLevel)
        current_index = level_values.index(self)
        next_index = current_index + 1
        if next_index < len(level_values):
            return level_values[next_index]
        return None


class TargetFilter(InputObjectType):
    label = Argument(NodeType)
    property_key = String()
    property_value = String()
    comparison = String()
    level = Argument(NodeLevel)


class Relationship(ObjectType):
    """Represents a directed relationship between two nodes in a graph."""

    id = String(required=True)
    source = Field(lambda: Node)
    target = Field(lambda: Node)
    label = String()
    properties = Field(JSONString)

    def __init__(self, id=None, **kwargs):
        super().__init__(id=id or str(uuid.uuid4()), **kwargs)


class Node(ObjectType):
    """Represents a node in a graph with associated properties."""

    id = String(required=True)
    label = Field(NodeType)
    properties = Field(JSONString)
    relationships = List(
        lambda: Relationship,
        label=Argument(String),
        property_key=Argument(String),
        property_value=Argument(String),
        comparison=Argument(String),
        target_filter=Argument(
            TargetFilter
        ),  # New argument to filter by target properties
    )
    level = Field(NodeLevel)

    def __init__(self, id=None, **kwargs):
        super().__init__(id=id or str(uuid.uuid4()), **kwargs)

    def resolve_relationships(
        self,
        info,
        label=None,
        property_key=None,
        property_value=None,
        comparison=None,
        target_filter=None,
    ):
        relationships = info.context.get("relationships", [])
        filtered_relationships = [
            rel for rel in relationships if rel.source.id == self.id
        ]

        if label:
            filtered_relationships = [
                rel for rel in filtered_relationships if rel.label == label
            ]

        if property_key and property_value:
            if comparison == "gt":
                filtered_relationships = [
                    rel
                    for rel in filtered_relationships
                    if rel.properties
                    and property_key in rel.properties
                    and float(rel.properties[property_key]) > float(property_value)
                ]
            elif comparison == "lt":
                filtered_relationships = [
                    rel
                    for rel in filtered_relationships
                    if rel.properties
                    and property_key in rel.properties
                    and float(rel.properties[property_key]) < float(property_value)
                ]
            elif comparison == "eq":
                filtered_relationships = [
                    rel
                    for rel in filtered_relationships
                    if rel.properties
                    and property_key in rel.properties
                    and rel.properties[property_key] == property_value
                ]
            else:
                raise ValueError(f"Invalid comparison operator: {comparison}")

        if target_filter:
            target_label = target_filter.get("label")
            target_property_key = target_filter.get("property_key")
            target_property_value = target_filter.get("property_value")
            target_comparison = target_filter.get("comparison")
            target_level = target_filter.get("level")

            filtered_relationships = [
                rel
                for rel in filtered_relationships
                if (
                    (not target_label or rel.target.label == target_label)
                    and (
                        not target_property_key
                        or (
                            target_property_key in rel.target.properties
                            and (
                                (
                                    target_comparison == "gt"
                                    and float(
                                        rel.target.properties[target_property_key]
                                    )
                                    > float(target_property_value)
                                )
                                or (
                                    target_comparison == "lt"
                                    and float(
                                        rel.target.properties[target_property_key]
                                    )
                                    < float(target_property_value)
                                )
                                or (
                                    target_comparison == "eq"
                                    and rel.target.properties[target_property_key]
                                    == target_property_value
                                )
                            )
                        )
                    )
                    and (not target_level or rel.target.level == target_level)
                )
            ]

        return filtered_relationships


class Query(ObjectType):
    filter_nodes = List(
        Node,
        ids=Argument(List(String)),
        label=Argument(NodeType),
        level=Argument(NodeLevel),
        property_key=Argument(String),
        property_value=Argument(String),
    )

    leaf_nodes = List(Node, id=Argument(String))

    @staticmethod
    def resolve_leaf_nodes(parent, info, id):
        def get_all_leaf_nodes(node):
            leaf_nodes = []
            next_level = node.level.next_level()
            child_relationships = []
            if node.relationships:
                child_relationships = [
                    relationship
                    for relationship in node.relationships
                    if relationship.label == "child"
                    and relationship.target.level == next_level
                    and relationship.source.id == node.id
                ]

            if not child_relationships:
                return [node]

            for relationship in child_relationships:
                leaf_nodes.extend(get_all_leaf_nodes(relationship.target))

            return leaf_nodes

        nodes = info.context.get("nodes", [])
        node = [node for node in nodes if node.id == id]
        if node:
            node = node[0]
            return get_all_leaf_nodes(node)
        else:
            return []

    @staticmethod
    def resolve_filter_nodes(
        parent,
        info,
        ids=None,
        label=None,
        level=None,
        property_key=None,
        property_value=None,
    ):
        nodes = info.context.get("nodes", [])
        filtered_nodes = nodes

        if ids:
            if isinstance(ids, list):
                filtered_nodes = [node for node in nodes if node.id in ids]
            else:
                filtered_nodes = [node for node in nodes if node.id == id]

        if label:
            filtered_nodes = [node for node in nodes if node.label == label]

        if level:
            filtered_nodes = [node for node in nodes if node.level.name == level.name]

        if property_key and property_value:
            filtered_nodes = [
                node
                for node in filtered_nodes
                if node.properties
                and node.properties.get(property_key) == property_value
            ]
        return filtered_nodes


schema = Schema(query=Query)
